/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * atm_iface.c: our interface to the ATM routines
 *
 * 29Oct92-20Nov92, mjz - old stuff
 * July-Oct 93, mjz - new stuff (Jose's interface)
 * 11 Nov 93 (mjz): switch to using reply port for requests
 * 11 Dec 93 (mjz): update for signalling interface.
 * 13 Jun 84 (mjz): update for >= jun 10 kernel (nw_endpoint_allocate changed).
 */
#include <stdio.h>
#include <cthreads.h>
#include "pbuf.h"		/* incl sys/types, comm, requires cthreads. */
#include <sys/uio.h>		/* for iov's */
#include <sys/socket.h>		/* for struct msghdr */
/*#include "atm_map.h"*/		/* the new ATM interface */
#include <nw.h>			/* fucking things moved */
#include "midway-dbg.h"		/* WIRE_DA_THREADS, etc. */
#include <flow_times.h>		/* gcc has problems with "" */
#include "comm.h"		/* damn.  did not want to include this. */
#include "atm_iface.h"		/* external fn definitions */
#include <strings.h>		/* for index() */

#define ATM_MAX_RCV_ERR		50 /* tolerate N successive read errors */

#if PREVIOUS_ATM_KERNEL
#define JCBKERNEL		940301 /* more or less date of kernel */
#else
/* current ATM kernel */
#define JCBKERNEL		940610
#endif

/* max atm message len is 64k, but realistically it's 12 'cause of rcv fifo */
#define MAX_ATM_SIZE		12*1024
#define MAX_ATM_MSG_SIZE       (mid_max_pbuf_size - sizeof(u_long))

#define MIDWAY_ATM_BANDWIDTH	10 /* Kbps, bogus */
#define MIDWAY_ATM_BANDWIDTH_MIN 1 /* ditto */

/*
 * ATM host is called via an "atm address".  There are two endpoints
 * created on each host, a calling endpoint and a destination endpoint.
 *
 * We are going to use sequenced packet protocol for each line.  There are 
 * 6 pipes and 8 lines for four machines.
 *
 * the addrmap is indexed by host #, so to send to host 1, you will
 * use the particular endpoint going towards that host on this machine.
 * Because we are statically setting up, the endpoint #'s have global
 * significance.
 *
 */

#define N_HSTS 8

/* following magic # from Jose is switch dependent (he gave it as the
   integer 4044620976) */
#define PEER_ADDR2	0xf11404b0


typedef struct {
  nw_peer_s	peer;		/* nw interface peer info: requests go here */
  nw_peer_s	rpc;		/* rpc's originate from here */
  nw_peer_s	meta;		/* except ones for metadata, which use this */
  int		port;		/* port on switch */
} atm_map_t;
atm_map_t *midway_atm_map;	/* map host# to atm endpoints & addrs */

int midway_atm_ep_num = 0;	/* # of EPs in... */
nw_ep midway_atm_ep_array[N_HSTS];
nw_ep midway_atm_ep_array_meta[N_HSTS];
nw_ep midway_atm_ep_array_rpc[N_HSTS];

u_int midway_atm_rcv_bufsize = 64*1024;	/* 64 K */ /* per endpoint */
u_int midway_atm_rcv_lg_bufsize = 64*1024; /* 128 K */ /*bug at 128*/

/*
 * for midway_atm_loop
 */
int midway_atm_loop_spins = 10; /* want about context switch time, I guess */
int midway_atm_vol_ctxsw = 0;	/* # of cthread_yeild() calls */
int midway_atm_ctx_spins = 0; /* max # of spins before vol ctx sw */
int midway_atm_nopack = 0;	/* set to zero when packet arrives */
mutex_t midway_atm_getlock;	/* lock get_message calls for plan C */
int midway_atm_rcv_spureous = 0; /* # of wakeups w/o anything to do */
int midway_atm_rcv_wakeups = 0;  /* # of wakeups with something to do */

int midway_atm_rcv_errors = 0;
int midway_atm_rcv_packets = 0;

int midway_atm_snd_overruns = 0; /* number of NW_OVERRUNs encountered */
int midway_atm_snd_maxspins = 0; /* times to spin to overcome overrun */
int midway_atm_snd_totspins = 0; /* total # of spins */
int midway_atm_snd_spin_max = 10000000; /* upper bound on maxspins */
int midway_atm_snd_peer = 0;	/* # sent out peer port */
int midway_atm_snd_rpc = 0;	/* # sent out rpc port */
int midway_atm_snd_meta = 0;	/* # sent out meta port */

/*
 * rcv plan b: keep a q of pbufs, hand off to rcv calls.
 */
pbufq_t midway_atm_buffer_queue;
pbuf_q midway_atm_bufq = &midway_atm_buffer_queue;
mutex_t midway_atm_lock;
condition_t midway_atm_c;
int midway_atm_queued;		/* actual number on Q. */
int midway_atm_upcall_queued;	/* so can tell if something q'd */
/* following assumes constant pool of receivers! - 14dec92 mjz */
int midway_atm_rcv_active = 0;	/* # of rcv's that deq'd, & not waiting. */

/*
 * rcv plan c: try not to context switch, but use plan b as default
 */
mutex_t midway_atm_fastlock;
condition_t midway_atm_fastcond;
int midway_atm_wantreply_queued = 0;
int midway_atm_wantreply_active = 0;
int midway_atm_rcv_rpc = 0;	/* # received via wantreply */
int midway_atm_rcv_rpc_punt = 0; /* # received but filter can't process */

int midway_atm_filter_succ = 0;	/* note successes */
int midway_atm_filter_mumiss = 0;
int midway_atm_filter_muhit = 0;
int midway_atm_locker = 0;
int midway_atm_filter_used = 0;

/*
 * rcv plan d: allow application to poll too (these are just counters)
 */
int midway_atm_poll_filter = 0;
int midway_atm_poll_mumiss = 0;
int midway_atm_poll_muhit = 0;
int midway_atm_poll_succ = 0;
int midway_atm_poll_pkts = 0;

/*
 * rcv plan e: interrupts
 */
int midway_atm_intr_filter = 0;
int midway_atm_intr_mumiss = 0;
int midway_atm_intr_muhit = 0;
int midway_atm_intr_succ = 0;
int midway_atm_intr_pkts = 0;

int midway_atm_intr_waitsig = 0; /* # signals done in midway_atm_willblock() */
int midway_atm_intr_waitnum = 0; /* like next, except never decreases */
int midway_atm_intr_waiting = 0; /* # threads in willblock */
mutex_t midway_atm_intr_wait_lock; /* protect waiting */

/*
 * externals
 */
u_long Comm_index;		/* our host number */
u_long Comm_hnum;		/* total # */


/* forward */


/* set up peer struct, allocate endpoints, and make connection */
int
midway_atm_allocate(int port, /* port on switch 0-7 */
		    nw_ep remote_ep,
		    nw_ep local_ep,
		    nw_peer_t peer,
		    int bufsize)
{
  nw_result	rc;

  /* rem_addr in peer now set up already. */
  peer->local_ep = local_ep;
  peer->remote_ep = remote_ep;

  /* open sending side */
  rc = nw_endpoint_allocate(
			    &local_ep,
			    NW_SEQ_PACKET, /* use Sequenced Packet Proto */
			    NW_DUPLEX, /* SPP's are duplex */
			    NW_AUTO_ACCEPT,
#if JCBKERNEL >= 940610
			    NW_NO_PMODE,
#endif
			    bufsize
			    );
  if (rc != NW_SUCCESS) {
    fprintf(stderr, "nw_endpoint_allocate source, %d to port %d (ep %d): %s\n",
	    Comm_index, port, local_ep, nw_result_char(rc));
    botch("nw_endpoint_allocate");
    return((int) rc);
  }
  return(0);
}
int
midway_atm_connect(int port, /* port on switch 0-7 */
		   nw_ep remote_ep,
		   nw_ep local_ep,
		   nw_peer_t peer,
		   int bufsize)	/* unused */
{
  nw_result	rc;
  int		bw;

  /* make connection */
  bw = MIDWAY_ATM_BANDWIDTH;	/* bogus */
  rc = nw_connection_open(
			  local_ep,
			  peer->rem_addr_1,
			  peer->rem_addr_2,
			  remote_ep,
			  &bw,
			  MIDWAY_ATM_BANDWIDTH_MIN
			  );
  if ((rc != NW_SUCCESS) && (rc != NW_ALREADY_CONNECTED) &&
      (rc != NW_BAD_ADDRESS)) {
    fprintf(stderr, "nw_connection_open, %d [ep %d] to port %d [ep %d] fails rc %d (%s)\n",
	    Comm_index, local_ep, port, remote_ep, rc, nw_result_char(rc));
    fprintf(stderr, "nw_connection_open, %d rem_addr_1 0x%08x rem_addr_2 0x%08x\n",
	    Comm_index, peer->rem_addr_1, peer->rem_addr_2);
    /* hmm... check status too, please. */
    {
      nw_state s;
      nw_peer_s peer;
      nw_result src = rc;
      rc = nw_endpoint_status(local_ep,
			      &s, &peer);
      if (rc != NW_SUCCESS) {
	fprintf(stderr, "nw_connection_open: endpoint status after open error returns %d (%s)\n",
		rc, nw_result_char(rc));
	botch("nw_endpoint_status");
      } else {
	fprintf(stderr, "nw_connection_open: %d open err, ep %d state %d (%s)\n",
		Comm_index, local_ep, s, nw_state_char(s));
	fprintf(stderr, "nw_connection_open: %d peer a1 0x%08x a2 0x%08x rem %d loc %d\n",
		Comm_index,
		peer.rem_addr_1, peer.rem_addr_2, peer.remote_ep, 
		peer.local_ep);
	rc = src;		/* reset for below */
      }
    }
    /* for now, just assume NW_BAD_ADDRESS is OK, and the other side
       has not cranked yet.  Also leave in debugging.  This will have
       to go away before real work gets done. */
    if ((rc != NW_BAD_ADDRESS) && (rc != NW_ALREADY_CONNECTED)){
      botch("nw_connection_open");
      return((int)rc);
    }
  }
  if (IS_DBG_COMM_ATM && (rc == NW_SUCCESS)) {	/* move under DBG when faster  */
    fprintf(stderr, "nw_conn_open: host %d [ep %d] to port %d [ep %d]\n",
	    Comm_index, local_ep, port, remote_ep);
  }
  if (IS_DBG_COMM_ATM &&( bw != MIDWAY_ATM_BANDWIDTH)) {
    /* current signalling version 12/15/93 does not return = */
    fprintf(stderr, "nw_conn_open: bandwidth %d != specified %d (min %d)\n",
	    bw, MIDWAY_ATM_BANDWIDTH, MIDWAY_ATM_BANDWIDTH_MIN);
  }
  /*
   * NOTE: could now pass allocation information in host_op message,
   * and not have to statically define any of the numbers.  do that
   * later?  (could have > 1 person on ATM, then) (but would be timesharing
   * anyhow, so we would probably not want to do that...) 12dec93 mjz
   */
  return(0);
} /* midway_atm_allocate_connect */
/*
 * open
 */
int
midway_atm_open(char *hlist)
{
  kern_return_t	krc;
  int		i, j;
  nw_result	rc;
  char		*ptr;
  char		*p;
  char		*nxt_name;
  int		nxt_len;
  nw_ep		local_ep;	/* ep used for local */
  int		local_port;	/* local port on switch */
  char		*nxtarg(char **p, char*brk); /* in libcs - a cmu-ism */
  char		save;
  nw_address_s	addr = {0};		/* used for lookup */

  /* there is no signalling protocol.  All our addresses are hard coded */
  if (Comm_hnum > N_HSTS) {
    fprintf(stderr, "%d: Comm_hnum %d > N_HSTS %d, quitting.\n",
	    Comm_index, Comm_hnum, N_HSTS);
    exit(1);
  }

  /*
   * we need to set up the atm_map, the mapping of host # to endpoint
   * information.
   * 	- using the hlist, figure out what ports the various hosts are on.
   *	- set up the endpoints
   */
  midway_atm_map = (atm_map_t *) CALLOC(Comm_hnum+1/*IO 0*/, sizeof(atm_map_t));
  if (midway_atm_map == 0)
    botch("atm_iface allocation of atm_map");

  ptr = hlist;
  for (i=1; i<=Comm_hnum; i++) { /* index origin 0 */
    nxt_name = nxtarg(&ptr, " 	"/*sp,tab*/);
    /* ptr is set following the break char (any others if multiple?) */
    if (*nxt_name == '\0') {
      fprintf(stderr,"midway_atm_open: ran out of hlist before Comm_hnum! [i=%d, hnum=%d]\n",
	      i, Comm_hnum);
      exit(1);
    }

    /* find name in using nw_lookup.  The name has no trailing bah.cs.cmu.edu*/
    save = *ptr;		/* save */
    *ptr = '\0';		/* ensure null */
    p = index(nxt_name,'.');	/* find 1st . in "mc20.mach.cs.cmu.edu" */
    *ptr = save;		/* reset */
    if (p == NULL)
      p = ptr-1;		/* set to 1 before end. */
    /* set up addr struct for lookup */
    save = *p;
    *p = '\0';
    strncpy(addr.name, nxt_name, sizeof(addr.name)-1);/*-1: always null term */
/*    *p = save;*/
#if 0
    fprintf(stderr, "%d going to look up '%s'\n", 
	    Comm_index, addr.name);/*XXX*/
#endif
    /* Don't forget to clear the IP address, or it will take that in
       preference to the specified name.  I.E. zero out the other fields! */
    addr.nw_addr_1 = NW_DEVICE_ENCODE(0,NW_TCA100_1); /* 1st ATM card */
    addr.nw_addr_2 = 0;
    addr.ip_addr = 0;
    rc = nw_lookup(NW_HOST_ADDRESS_LOOKUP, (int*)(&addr));
    if (rc != NW_SUCCESS) {
      fprintf(stderr, "%d nw_lookup %s: %s\n",
	      Comm_index, nxt_name, nw_result_char(rc));
      botch("nw_lookup");
      return((int) rc);
    }
    /* after lookup, port appears to be encoded in hi order byte */
    /* also, if nw2 is not setup, it appears to have grabbed the loopback.
       this seems to occur when passing a bogus name in... trap it here */
    /* yeah, but might want a 0 if we run back-to-back.  Try matching names */
#if 0
    fprintf(stderr,"conn open %d addr.name '%s' nxt_name '%s', sizeof(addr.name)-1 %d cmp %d\n",
	    Comm_index, addr.name, nxt_name, sizeof(addr.name)-1); /*XXX*/
#endif
    if (strncmp(addr.name, nxt_name,sizeof(addr.name)-1)) {
      *ptr = '\0';		/* delimit for print *//*DON'T NEED,SEENEXT*/
      fprintf(stderr,"midway_atm_open: %s not in atm configuration.\n",
	      nxt_name);
      exit(1);
    }
    *p = save;			/* this is why DON'T NEED above */

    midway_atm_map[i].port = 
      (unsigned)(NW_ADDRESS_DECODE(addr.nw_addr_1)) >> 24;
    /* toss IP & returned name, but save "network" address */
    midway_atm_map[i].meta.rem_addr_1 =
      midway_atm_map[i].rpc.rem_addr_1 =
	midway_atm_map[i].peer.rem_addr_1 = addr.nw_addr_1;
    midway_atm_map[i].meta.rem_addr_2 =
      midway_atm_map[i].rpc.rem_addr_2 =
	midway_atm_map[i].peer.rem_addr_2 = addr.nw_addr_2;
    if (i == Comm_index) {
      char smsg[255];
      sprintf(smsg,"ATM open: I am %s (%s) and am on the switch at port %d.",
	      nxt_name, addr.name, midway_atm_map[i].port);
      StatusMsg(smsg);
    }
  }

  /* remote ep for all is based on local atm port! */
  local_port = midway_atm_map[Comm_index].port;

  for (i = 1; i <= Comm_hnum; i++) {
    if (i != Comm_index) {	/* for all but ourselves */
      int remote_port = midway_atm_map[i].port;

      local_ep = 11 + remote_port;
      midway_atm_allocate(
			  remote_port,			/* rem port*/
			  11+local_port,		/* remote_ep */
			  local_ep,			/* local_ep */
			  &(midway_atm_map[i].peer),	/* peer strt */
			  midway_atm_rcv_lg_bufsize
			  );

      /* keep an array of endpoints for select operation */
      midway_atm_ep_array[midway_atm_ep_num] = local_ep;


      /* now do one used for an RPC */
      midway_atm_allocate(
			  remote_port,
			  21+local_port,
			  21+remote_port,
			  &(midway_atm_map[i].rpc),
			  midway_atm_rcv_bufsize
			  );
      midway_atm_ep_array_rpc[midway_atm_ep_num] = 21+remote_port; /*local_ep*/

      /* now do one used for metadata */
      midway_atm_allocate(
			  remote_port,
			  1+local_port,
			  1+remote_port,
			  &(midway_atm_map[i].meta),
			  midway_atm_rcv_bufsize
			  );
      midway_atm_ep_array_meta[midway_atm_ep_num] = 1+remote_port; /*local_ep*/

      midway_atm_ep_num++;
				  
    } /* i!=Comm_index */
  } /* for each destination host */

  /* open as second phase, so we don't deadlock */
  for (i = 1; i <= Comm_hnum; i++) {
    if (i != Comm_index) {	/* for all but ourselves */
      int remote_port = midway_atm_map[i].port;

      midway_atm_connect(
			  remote_port,			/* rem port*/
			  midway_atm_map[i].peer.remote_ep,/* remote_ep */
			  midway_atm_map[i].peer.local_ep, /* local_ep */
			  &(midway_atm_map[i].peer),	/* peer strt */
			  midway_atm_rcv_lg_bufsize
			  );


      /* now do one used for an RPC */
      midway_atm_connect(
			  remote_port,
			  midway_atm_map[i].rpc.remote_ep,
			  midway_atm_map[i].rpc.local_ep,
			  &(midway_atm_map[i].rpc),
			  midway_atm_rcv_bufsize
			  );

      /* now do one used for metadata */
      midway_atm_connect(
			  remote_port,
			  midway_atm_map[i].meta.remote_ep,
			  midway_atm_map[i].meta.local_ep,
			  &(midway_atm_map[i].meta),
			  midway_atm_rcv_bufsize
			  );
    } /* i!=Comm_index */
  } /* for each destination host */

  midway_atm_lock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(midway_atm_lock,"midway_atm_lock");
  midway_atm_c = condition_alloc();
  mid_pbuf_qinit(midway_atm_bufq);
  midway_atm_fastlock = mutex_alloc();
  MIDWAY_MUTEX_SET_NAME(midway_atm_fastlock,"midway_atm_fastlock");
  midway_atm_fastcond = condition_alloc();

  /* ready to receive */
  return(0);
} /* midway_atm_open */

/*
 * sendmsg
 */
#if FLOW_TIMES
unsigned int ft_send_time_prealloc;
unsigned int ft_send_time_alloc;
unsigned int ft_send_time_start; /* gak.  not thread safe */
unsigned int ft_send_time_end;	/* ditto */
#endif
int
midway_atm_sendmsg(
		   int s,	/* socket, not used */
		   struct msghdr *msgh, /* gather array */
		   int flags,
		   int dest	/* host # sending to */
		   )
{
  nw_buffer_t nwbuf;
  int i;
  char *c;
  nw_result rc;
  int len = 0;
  pbuf pb;
  int spins;

  /* compute total length */
  for (i=0; i<msgh->msg_iovlen; i++) {
    len += msgh->msg_iov[i].iov_len;
  }

  if (len > MAX_ATM_MSG_SIZE) {
    fprintf(stderr, "atm_sendmsg: message too big: %lu > %lu dropping.\n",
	    len, MAX_ATM_MSG_SIZE);
    botch("midway_atm_sendmsg");
  }

  /* gather */
#define GATHER(dest) \
  for (c=(dest), i=0; i<msgh->msg_iovlen; i++) {\
    bcopy(msgh->msg_iov[i].iov_base, c, msgh->msg_iov[i].iov_len);\
    c += msgh->msg_iov[i].iov_len;\
  }

  /* cheap loopback hack.  Only inv or inv reply are sent this way,
     so just bcopy into a new pbuf.  We must do this because
     the switch is not configured to send to ourselves.  This could
     be done, but perhaps it is better to do this here anyway.
     4 nov 92 mjz
   */
  if (dest == Comm_index) {	/* short circuit */
    pb = mid_pbuf_get();
#if FLOW_TIMES
    /* see upcall */
    GATHER((pb+8));
    *((u_int *)(pb+4)) = FT_READ_TIMER;
#else
    GATHER((pb+sizeof(u_long)));
#endif
    *((u_long *)pb) = (u_long) len;
    mutex_lock(midway_atm_lock);
    mid_pbuf_enqueue(pb, midway_atm_bufq);
    midway_atm_queued++;
    mutex_unlock(midway_atm_lock);
    condition_signal(midway_atm_c);
  } else {
    /* see what stream this is supposed to be in (req, repl, or metadata) */
    /* this is a hack! */
    nw_peer_t x;
    COMMMSGP cm = (COMMMSGP)(msgh->msg_iov[0].iov_base); /* better be, anyway */
    register u_char cm_type = cm->type;

    if (((cm_type & M_OP_BITS) == M_ALLOCATE) ||
	((cm_type & M_OP_BITS) == M_ALLOC_INFO)) {	/* metadata */
      x = &(midway_atm_map[dest].meta);
      midway_atm_snd_meta++;
    }
    else if ((cm_type == (M_LOCK_OP|M_REPLY_BIT)) ||
	     (cm_type == (M_LOCK_INV_OP|M_REPLY_BIT)) ||
	     (cm_type == (M_INVALIDATE_OP|M_REPLY_BIT))){
      x = &(midway_atm_map[dest].rpc);
      midway_atm_snd_rpc++;
    }
    else {			/* must be request */
      x = &(midway_atm_map[dest].peer);
      midway_atm_snd_peer++;
    }

#if FLOW_TIMES
    ft_send_time_prealloc = FT_READ_TIMER;
#endif

    /* gotta copy into a kernel buffer for now (yuck) */
    nwbuf = nw_buffer_allocate(x->local_ep,
			       sizeof(nw_buffer_s) + len);
    /* don't need the following for SPP sends? */
    nwbuf->peer = *x;
    nwbuf->block_offset = sizeof(nw_buffer_s);
    nwbuf->block_length = len;
    nwbuf->block_deallocate = TRUE;
    nwbuf->buf_next = NULL;	/* no next! */

#if FLOW_TIMES
    ft_send_time_alloc = FT_READ_TIMER;
#endif

    GATHER(((char *)(nwbuf+1)));

    spins = 0;
    do {
#if FLOW_TIMES
      ft_send_time_start = FT_READ_TIMER; /* gak not thread safe. See l_Xfer*/
#endif
      rc = nw_send(x->local_ep,
		   nwbuf,
		   NW_NORMAL
		   );
#if FLOW_TIMES
      ft_send_time_end = FT_READ_TIMER;
#endif
      if ((rc != NW_SUCCESS) && (rc != NW_OVERRUN))
	fprintf(stderr, "nw_send from %d to %d: %s\n",
	      Comm_index, dest, nw_result_char(rc));

      /* try yeilding if nw_send is an overrun... probably true that
	 something else needs to happen (like a metadata message) to
	 have any movement on this VC */
#if 0
      if (rc == NW_OVERRUN)
#endif
	cthread_yield();
    } while(
#if (BENCHMARK != 0)
	    (rc == NW_OVERRUN) && 
#else
	    (rc != NW_SUCCESS) &&
#endif
	    (++spins < midway_atm_snd_spin_max));
    if (spins > 1) {
      if (spins > midway_atm_snd_maxspins)
	midway_atm_snd_maxspins = spins;
      midway_atm_snd_totspins += spins;
      /* worry abouy 1 vs 2 ?  What's in a spin? */
      midway_atm_snd_overruns++; /* must have, if spun */
      /* gosh, has buffer been deallocated if error? */
    }
    if (rc != NW_SUCCESS) {
      fprintf(stderr, "nw_send from %d to %d ep %d: %s\n",
	      Comm_index, dest, x->local_ep, nw_result_char(rc));
      botch("nw_send(sendmsg)");
    }
  }
  return(len);
} /* midway_atm_sendmsg */

/*
 * sendto
 */
int
midway_atm_sendto(
		  int s,	/* socket, not used */
		  char * buf,	/* thing to be sent */
		  int len,	/* size */
		  int flags,	/* not used */
		  struct sockaddr *to, /* not used */
		  int tolen,	/* not used */
		  int dest
		   )
{
  nw_result rc;
  pbuf pb;
  nw_buffer_t nwbuf;
  int spins;

  if (len > MAX_ATM_MSG_SIZE) {
    fprintf(stderr, "atm_sendto: message too big: %lu > %lu dropping.\n",
	    len, MAX_ATM_MSG_SIZE);
    botch("midway_atm_sendto");
  }

  /* cheap loopback hack, part II.  See sendmsg above. */
  if (dest == Comm_index) {	/* short circuit */
    pb = mid_pbuf_get();
#if FLOW_TIMES
    /* see upcall */
    bcopy(buf, pb+8, len);
    *((u_int *)(pb+4)) = FT_READ_TIMER;
#else
    bcopy(buf, pb+sizeof(u_long), len);
#endif
    *((u_long *)pb) = (u_long) len;
    mutex_lock(midway_atm_lock);
    mid_pbuf_enqueue(pb, midway_atm_bufq);
    midway_atm_queued++;
    mutex_unlock(midway_atm_lock);
    condition_signal(midway_atm_c);
  } else {
    /* see what stream this is supposed to be in (req, repl, or metadata) */
    /* this is a hack! */
    nw_peer_t x;
    COMMMSGP cm = (COMMMSGP) buf; /* better be, anyway */
    register u_char cm_type = cm->type;

    if (((cm_type & M_OP_BITS) == M_ALLOCATE) ||
	((cm_type & M_OP_BITS) == M_ALLOC_INFO)) {	/* metadata */
      x = &(midway_atm_map[dest].meta);
      midway_atm_snd_meta++;
    }
    else if ((cm_type == (M_LOCK_OP|M_REPLY_BIT)) ||
	     (cm_type == (M_LOCK_INV_OP|M_REPLY_BIT)) ||
	     (cm_type == (M_INVALIDATE_OP|M_REPLY_BIT))){
      x = &(midway_atm_map[dest].rpc);
      midway_atm_snd_rpc++;
    }
    else {			/* must be request */
      x = &(midway_atm_map[dest].peer);
      midway_atm_snd_peer++;
    }

#if FLOW_TIMES
    ft_send_time_prealloc = FT_READ_TIMER;
#endif

    /* gotta copy into a kernel buffer for now (yuck) */
    nwbuf = nw_buffer_allocate(x->local_ep,
			       sizeof(nw_buffer_s) + len);
    /* don't need the following for SPP sends? */
    nwbuf->peer = *x;
    nwbuf->block_offset = sizeof(nw_buffer_s);
    nwbuf->block_length = len;
    nwbuf->block_deallocate = TRUE;
    nwbuf->buf_next = NULL;

#if FLOW_TIMES
    ft_send_time_alloc = FT_READ_TIMER;
#endif

    bcopy(buf, (char *)(nwbuf+1), len);

    spins = 0;
    do {
#if FLOW_TIMES
      ft_send_time_start = FT_READ_TIMER;
#endif
      rc = nw_send(x->local_ep,
		   nwbuf,
		   NW_NORMAL
		   );

#if FLOW_TIMES
      ft_send_time_end = FT_READ_TIMER;
#endif

      if ((rc != NW_SUCCESS) && (rc != NW_OVERRUN))
	fprintf(stderr, "nw_send from %d to %d: %s\n",
	      Comm_index, dest, nw_result_char(rc));

      /* try yeilding if nw_send is an overrun... probably true that
	 something else needs to happen (like a metadata message) to
	 have any movement on this VC */
#if 0
      if (rc == NW_OVERRUN)
#endif
	cthread_yield();

    } while(
#if (BENCHMARK != 0)
	    (rc == NW_OVERRUN) && 
#else
	    (rc != NW_SUCCESS) &&
#endif
	    (++spins < midway_atm_snd_spin_max));

    if (spins > 1) {
      if (spins > midway_atm_snd_maxspins)
	midway_atm_snd_maxspins = spins;
      midway_atm_snd_totspins += spins;
      /* worry abouy 1 vs 2 ?  What's in a spin? */
      midway_atm_snd_overruns++; /* must have, if spun */
      /* gosh, has buffer been deallocated if error? */
    }

    if (rc != NW_SUCCESS) {
      fprintf(stderr, "nw_send from %d to %d: %s\n",
	      Comm_index, dest, nw_result_char(rc));
      botch("nw_send(sendto)");
    }
  }
  return(len);
} /* midway_atm_sendto */

/*
 * recvfrom; nah, don't bother.  We want to give it pbufs.
 */
int midway_atm_rcv_onq = 0;	/* for %^& loopback */
cthread_t midway_atm_rcv_meta = 0; /* another hack */
#if FLOW_TIMES
unsigned int ft_rcv_time_start;
unsigned int ft_rcv_time_copy; /* gak.  not thread safe */
unsigned int ft_rcv_time_rel;	/* ditto */
#endif
int
midway_atm_rcv(
	       pbuf *buf, 
	       int *len, 
	       u_int *stime,
	       struct sockaddr *sin
	       )
{
  pbufe_t *qtop;
  nw_buffer_t nwbuf;
  nw_result rc;
  int errors = 0;
  int meta;
  nw_ep *wait_ep_array;

  mutex_lock(midway_atm_lock);

  /* only have one thread wait on metadata, dynamically assigned... */
  if (midway_atm_rcv_meta == 0) {
    midway_atm_rcv_meta = cthread_self(); /* barf-o-matic.  true if not wired
					   down??  I hope so. */
  }

  meta = (midway_atm_rcv_meta == cthread_self());

  if (midway_atm_rcv_onq || meta) {
    /*
     * someone waiting for queued (loopback) packets.  We will grab from net.
     */
    mutex_unlock(midway_atm_lock);

    if (meta) {
      /* ADSF.  Even if have separate ep's per host, need to wait on multiple
	 if multiple receivers.  Perhaps do an inverted multicast for rcv?
	 ACK. Is this even implemented? */
      /* really should be a pointer to an ep_array */
      wait_ep_array = midway_atm_ep_array_meta;
    }
    else {
      wait_ep_array = midway_atm_ep_array;
    }

#define NW_WAIT_FOREVER		-1

    *buf = mid_pbuf_get();

    nwbuf = (nw_buffer_t) NW_BUFFER_ERROR;

#if FLOW_TIMES
    ft_rcv_time_start = FT_READ_TIMER;
    /* this means that there is loop overhead, etc in the timing... */
#endif

    while (nwbuf == (nw_buffer_t)NW_BUFFER_ERROR) {

#if 1
      nwbuf = nw_select(midway_atm_ep_num, wait_ep_array, 
			NW_WAIT_FOREVER);
#else
      /* TESTING: try recv instead of select to see if we can avoid problem */
      /* XXXX only can wait on one EP this way, however! */
      nwbuf = nw_receive(wait_ep_array[0], NW_WAIT_FOREVER);
#endif

      /* this one seems to return 2 sometimes??? */
      if ((nwbuf > (nw_buffer_t)0) && (nwbuf < (nw_buffer_t)0x1000)) {
	fprintf(stderr,"midway_atm_rcv nw_receive returned bogus address 0x%x\n",
		nwbuf);
	nwbuf = (nw_buffer_t) NW_BUFFER_ERROR; /* treat as buffer error */
      }
/* someday, #if (BENCHMARK == 0) */
      /*
       * Sanity checks on incoming buffer.  This usually only happens
       * in gdb, put perhaps also occasionally?
       */
      if ((nwbuf != 0) && (nwbuf != (nw_buffer_t)NW_BUFFER_ERROR)  &&
	  ((nwbuf->buf_used == 0) || 
	  (nwbuf->buf_next >= (nw_buffer_t) 0x80000000) ||
	  (nwbuf->block_offset == 0) || 
	  (nwbuf->block_length > mid_max_pbuf_size) ||
	  (nwbuf->block_length > nwbuf->buf_length))) {
	int is_remote = midway_uxpass_local();
	fprintf(stderr, "Got invalid buffer: @0x%x used=%d len=%d next=0x%x seq=%d offs=%d dealloc=%d datlen=%d peer.a1=0x%x peer.a2=0x%x peer.remote_ep=%d peer.local_ep=%d\n",
		nwbuf, nwbuf->buf_used, nwbuf->buf_length, nwbuf->buf_next,
		nwbuf->msg_seqno, nwbuf->block_offset, nwbuf->block_deallocate,
		nwbuf->block_length, nwbuf->peer.rem_addr_1,
		nwbuf->peer.rem_addr_2, nwbuf->peer.remote_ep,
		nwbuf->peer.local_ep);
	/* print out in hex, because we cannot do so in gdb with proj. buff. */
	{ int *f = (int *)nwbuf;
	  fprintf(stderr,"[%08lx.%08lx.%08lx.%08lx.%08lx.%08lx.%08lx]\n",
		  *f, *(f+1),*(f+2),*(f+3),*(f+4),*(f+5),*(f+6));
	}
	fprintf(stderr, "[size of nw_buffer_s: %d]\n", sizeof(nw_buffer_s));
	if (is_remote)
	  midway_uxpass_remote();
	nwbuf = (nw_buffer_t) NW_BUFFER_ERROR; /* treat as buffer error */
      } /* sanity */
/* someday, #endif / *~BENCHMARK* / */

      if (nwbuf == 0) {
	/* ??? timeout satisfied?  What about signal, or something? */
	fprintf(stderr,"midway_atm_rcv select TIMED OUT? on host %d.  Dying.\n",
		Comm_index);
	cthread_exit((void *)1);
      } /* timeout expired */

      else if (nwbuf == (nw_buffer_t) NW_BUFFER_ERROR) {
	/* hack, check 1st ep in ep list (fine for 2 procs!)*/
	if (!errors) {
	  nw_state s;
	  nw_peer_s peer;
	  rc = nw_endpoint_status(wait_ep_array[0],
				  &s, &peer);
	  if (rc != NW_SUCCESS) {
	    fprintf(stderr, "midway_atm_rcv: endpoint status after select error returns %d (%s)\n",
		    rc, nw_result_char(rc));
	    cthread_exit((void *) 1);
	  } else {
	    fprintf(stderr, "mw_atm_rcv: sel err, ep %d state %d (%s)\n",
		    wait_ep_array[0], s, nw_state_char(s));
	  }
	}
	errors++;
	if (errors > ATM_MAX_RCV_ERR) {
	  fprintf(stderr,"midway_atm_rcv host %d got %d successive buffer errors from nw_select, quitting.\n",
		  Comm_index, errors);
#if STOP_ON_ERROR
	  botch("too many select errors");
#else
	  cthread_exit((void *)1);
#endif
	}
      } /* buffer error */

    } /* while */

    midway_atm_rcv_errors += errors; /* note for posterity */



#if FLOW_TIMES
    *stime = FT_READ_TIMER;
    *buf = *buf + 8;		/* XXXX see free buffer below */
#else
    *buf = *buf + 4;		/* XXXX */
#endif

    /* copy to pbuf.  barf */
    /* possible to receive more than one? */
    *len = nwbuf->block_length;
    if (*len > mid_max_pbuf_size) { /* should not happen due to check above */
      *len = mid_max_pbuf_size;
    }
    bcopy(((char *)nwbuf)+nwbuf->block_offset, *buf, *len);

#if FLOW_TIMES
    ft_rcv_time_copy = FT_READ_TIMER; /* GAK.  see comm_rcv stuff */
#endif

    if (nwbuf->buf_next) {
      nw_buffer_t bufP,bufQ;
      int limit = 100;		/* bail after 100. */
      int is_remote = midway_uxpass_local();

      fprintf(stderr, "Got buffer w/buf_next: @0x%x used=%d len=%d next=0x%x seq=%d offs=%d dealloc=%d datlen=%d\n",
	      nwbuf, nwbuf->buf_used, nwbuf->buf_length, nwbuf->buf_next,
	      nwbuf->msg_seqno, nwbuf->block_offset, nwbuf->block_deallocate,
	      nwbuf->block_length);
      fprintf(stderr, "[size of nw_buffer_s: %d]\n", sizeof(nw_buffer_s));

      /* try deallocating all following, because just deallocating string
	 fails miserably. */
      bufP = nwbuf->buf_next;
      if ((bufP < (nw_buffer_t) 0x400000) || (bufP > (nw_buffer_t) 0x7fffffff)) {
	/* certain failure on MIPS */
	fprintf(stderr, "Invalid buf_next pointer: 0x%x, ignoring\n",
		bufP);
	bufP = 0;
      }
      nwbuf->buf_next = 0;
      for(limit=0; ((limit<100) && bufP); limit++) {
	fprintf(stderr, "Deallocating: @0x%x used=%d len=%d next=0x%x seq=%d offs=%d dealloc=%d datlen=%d\n",
		bufP, bufP->buf_used, bufP->buf_length, bufP->buf_next,
		bufP->msg_seqno, bufP->block_offset, bufP->block_deallocate,
		bufP->block_length);

	bufQ = bufP->buf_next;
	bufP->buf_next = 0;	/* hopefully this is what was causing probs */

	rc = nw_buffer_deallocate(nwbuf->peer.local_ep, nwbuf);
	if (rc != NW_SUCCESS) {
	  fprintf(stderr,"nw_buffer_deallocate failed, rc %d (%s)\n",
		  rc, nw_result_char(rc));
	  botch("nw_buffer_deallocate");
	}

	bufP = bufQ;
	if ((bufP < (nw_buffer_t)0x400000) || (bufP > (nw_buffer_t)0x7fffffff)) {
	  /* certain failure on MIPS */
	  fprintf(stderr, "Invalid buf_next pointer: 0x%x, ignoring\n",
		  bufP);
	  bufP = 0;
	}
      } /* for */
      if (is_remote)
	(void) midway_uxpass_remote();
    }
    rc = nw_buffer_deallocate(nwbuf->peer.local_ep, nwbuf);
    if (rc != NW_SUCCESS) {
      fprintf(stderr,"nw_buffer_deallocate failed, rc %d (%s)\n",
	      rc, nw_result_char(rc));
      botch("nw_buffer_deallocate");
    }
#if FLOW_TIMES
    ft_rcv_time_rel = FT_READ_TIMER;
#endif
  } /* receive from network */

  else {
    /*
     * No one waiting for queued (loopback) packets, we will check.
     */
    midway_atm_rcv_onq++;	/* note that we are on the job. */

    if (midway_atm_rcv_active > 0)
      midway_atm_rcv_active--;

    qtop = MID_PBUF_QTOP(midway_atm_bufq);
    while (qtop == 0) {
      condition_wait(midway_atm_c, midway_atm_lock);
      qtop = MID_PBUF_QTOP(midway_atm_bufq);
      if (qtop == 0)
	midway_atm_rcv_spureous++;
    }

    *buf = mid_pbuf_dequeue(midway_atm_bufq);
    midway_atm_queued--;

    midway_atm_rcv_active++;

    midway_atm_rcv_onq--;

    mutex_unlock(midway_atm_lock);

    *len = *((u_long *)(*buf));	/* grab length off top */
#if FLOW_TIMES
    *stime = *((u_int *)(*buf+4));
    *buf += 8;
#else
    *buf += sizeof(u_long);	/* bump pointer so at data */
#endif
    midway_atm_rcv_wakeups++;
  } /* no one waiting  for queued packets */

  midway_atm_rcv_packets++;

  return(*len);
} /* midway_atm_rcv */

/*
 * pbuf free for rcv supplied pbufs
 */
midway_atm_pbuf_free(pbuf buf)
{
#if FLOW_TIMES
  ((char *)buf) -= 8;
#else
  ((char *)buf) -= sizeof(u_long); /* MORE HACKS. DAMN. */
#endif
  return(mid_pbuf_free(buf));
} /* midway_atm_pbuf_free */


/*
 * wantreply: the current thread sent a request, and is looking for
 * a reply.  We will field the replies (so there is no other context
 * switching).  In case there are multiple requestors, we must
 * stay here until all packets have been received; if one is not
 * for us, shunt it to the general receiver. Yuck.
 */
int
midway_atm_wantreply(comm_filter filt, int cpti)
{
  pbuf pb;
  char *buf;
  int len; /* of msg in buf */
  nw_buffer_t nwbuf;
  nw_result rc;
  int errors = 0;
  int done = 0;

  mutex_lock(midway_atm_fastlock);
  /*
   * problem: if have more than one thread wait in ATM code, could get
   * the following scenario:
   *    A waits for ReplA
   *    B waits for ReplB
   *    ReplB arrives, and is serviced by A.  A punts this (?perhaps does it),
   *       and must go back for waiting for replA.
   *    ReplA arrives, and is serviced by B.  B punts this, but it can know
   *       that replB arrived, so it can exit.  However, poor A is still stuck
   *       waiting for ReplA, which has already arrived, and there is no way
   *       for us to break the wait.
   * solution: only have on thread ever wait; if it is satisfied, release
   *       an other waiter.  In addition need to kick off other waiters early?
   *       Problem: do not know when packet is finished processing.  Sigh.
   */
   if (midway_atm_wantreply_active) {
     while (midway_atm_wantreply_active) {
       midway_atm_wantreply_queued++;
       condition_wait(midway_atm_fastcond, midway_atm_fastlock);
       midway_atm_wantreply_queued--;
     }
     if (midAllpackets(cpti)) {
       condition_signal(midway_atm_fastcond); /* let another waiter go */
       mutex_unlock(midway_atm_fastlock);
       return(1); /* done! */
     }
   }
  midway_atm_wantreply_active++; /* note we will be waiting... */
  mutex_unlock(midway_atm_fastlock);

  pb = mid_pbuf_get();
  while(!done) {
    /*
     * receive from net
     */

    nwbuf = (nw_buffer_t) NW_BUFFER_ERROR;

#if FLOW_TIMES
    ft_rcv_time_start = FT_READ_TIMER;
#endif

    while (nwbuf == (nw_buffer_t)NW_BUFFER_ERROR) {

#if 1
      nwbuf = nw_select(midway_atm_ep_num, midway_atm_ep_array_rpc,
			NW_WAIT_FOREVER);
#else
      /* TESTING: try recv instead of select to see if we can avoid problem */
      /* XXXX only can wait on one EP this way, however! */
      nwbuf = nw_receive(midway_atm_ep_array_rpc[0], NW_WAIT_FOREVER);
#endif

      /* this one seems to return 2 sometimes??? */
      if ((nwbuf > (nw_buffer_t)0) && (nwbuf < (nw_buffer_t)0x1000)) {
	fprintf(stderr,"midway_atm_waitreply nw_receive returned bogus address 0x%x\n",
		nwbuf);
	nwbuf = (nw_buffer_t) NW_BUFFER_ERROR; /* treat as buffer error */
      }

      if (nwbuf == 0) {
	/* ??? timeout satisfied?  What about signal, or something? */
	fprintf(stderr,"midway_atm_waitreply select TIMED OUT? on host %d.  Dying.\n",
		Comm_index);
	botch("waitreply select timed out");
      } /* timeout expired */

      else if (nwbuf == (nw_buffer_t) NW_BUFFER_ERROR) {
	/* hack, check 1st ep in ep list (fine for 2 procs!)*/
	if (!errors) {
	  nw_state s;
	  nw_peer_s peer;
	  rc = nw_endpoint_status(midway_atm_ep_array_rpc[0],
				  &s, &peer);
	  if (rc != NW_SUCCESS) {
	    fprintf(stderr, "midway_atm_waitreply: endpoint status after select error returns %d (%s)\n",
		    rc, nw_result_char(rc));
	    botch("atm_waitreply, nw_endpoint_status fails");
	  } else {
	    fprintf(stderr, "mw_atm_waitreply: sel err, ep %d state %d (%s)\n",
		    midway_atm_ep_array_rpc[0], s, nw_state_char(s));
	  }
	}
	errors++;
	if (errors > ATM_MAX_RCV_ERR) {
	  fprintf(stderr,"midway_atm_waitreply host %d got %d successive buffer errors from nw_select, quitting.\n",
		  Comm_index, errors);
#if STOP_ON_ERROR
	  botch("atm_waitreply too many errs");
#else
	  cthread_exit((void *)1);
#endif
	}
      } /* buffer error */

    } /* while */

#if FLOW_TIMES
    *((u_int *)(pb+4)) = FT_READ_TIMER;
    buf = pb + 8;		/* XXXX see free buffer above */
#else
    buf = pb + 4;		/* XXXX */
#endif

    midway_atm_rcv_errors += errors; /* note for posterity */

    midway_atm_rcv_rpc++;

    /* copy to pbuf.  barf */
    /* possible to receive more than one? */
    len = nwbuf->block_length;
    if (len > mid_max_pbuf_size) {
      len = mid_max_pbuf_size;
    }
    bcopy(((char *)nwbuf)+nwbuf->block_offset, buf, len);

#if FLOW_TIMES
    ft_rcv_time_copy = FT_READ_TIMER; /* GAK.  see comm_rcv stuff */
#endif

/* someday, #if (BENCHMARK == 0) */
    /*
     * Sanity checks on incoming buffer.  This usually only happens
     * in gdb, put perhaps also occasionally?
     */
    if ((nwbuf->buf_used == 0) || 
	(nwbuf->buf_next >= (nw_buffer_t)0x80000000) ||
	(nwbuf->block_offset == 0) || 
	(nwbuf->block_length > mid_max_pbuf_size) ||
	(nwbuf->block_length > nwbuf->buf_length)) {
      int is_remote = midway_uxpass_local();
      fprintf(stderr, "WR.Got invalid buffer: @0x%x used=%d len=%d next=0x%x seq=%d offs=%d dealloc=%d datlen=%d peer.a1=0x%x peer.a2=0x%x peer.remote_ep=%d peer.local_ep=%d\n",
	      nwbuf, nwbuf->buf_used, nwbuf->buf_length, nwbuf->buf_next,
	      nwbuf->msg_seqno, nwbuf->block_offset, nwbuf->block_deallocate,
	      nwbuf->block_length, nwbuf->peer.rem_addr_1,
	      nwbuf->peer.rem_addr_2, nwbuf->peer.remote_ep,
	      nwbuf->peer.local_ep);
      fprintf(stderr, "WR.[size of nw_buffer_s: %d]\n", sizeof(nw_buffer_s));
      if (is_remote)
	midway_uxpass_remote();
      continue;			/* try receiving new packet */
    } /* sanity */
/* someday, #endif / *~BENCHMARK* / */

    if (nwbuf->buf_next) {
      nw_buffer_t bufP,bufQ;
      int limit = 100;		/* bail after 100. */
      int is_remote = midway_uxpass_local();

      fprintf(stderr, "WR.Got buffer w/buf_next: @0x%x used=%d len=%d next=0x%x seq=%d offs=%d dealloc=%d datlen=%d\n",
	      nwbuf, nwbuf->buf_used, nwbuf->buf_length, nwbuf->buf_next,
	      nwbuf->msg_seqno, nwbuf->block_offset, nwbuf->block_deallocate,
	      nwbuf->block_length);
      fprintf(stderr, "WR.[size of nw_buffer_s: %d]\n", sizeof(nw_buffer_s));

      /* try deallocating all following, because just deallocating string
	 fails miserably. */
      bufP = nwbuf->buf_next;
      if ((bufP < (nw_buffer_t)0x400000) || (bufP > (nw_buffer_t)0x7fffffff)) {
	/* certain failure on MIPS */
	fprintf(stderr, "WR.Invalid buf_next pointer: 0x%x, ignoring\n",
		bufP);
	bufP = 0;
      }
      nwbuf->buf_next = 0;
      for(limit=0; ((limit<100) && bufP); limit++) {
	fprintf(stderr, "WR.Deallocating: @0x%x used=%d len=%d next=0x%x seq=%d offs=%d dealloc=%d datlen=%d\n",
		bufP, bufP->buf_used, bufP->buf_length, bufP->buf_next,
		bufP->msg_seqno, bufP->block_offset, bufP->block_deallocate,
		bufP->block_length);

	bufQ = bufP->buf_next;
	bufP->buf_next = 0;	/* hopefully this is what was causing probs */

	rc = nw_buffer_deallocate(nwbuf->peer.local_ep, nwbuf);
	if (rc != NW_SUCCESS) {
	  fprintf(stderr,"WR.nw_buffer_deallocate failed, rc %d (%s)\n",
		  rc, nw_result_char(rc));
	  botch("nw_buffer_deallocate");
	}

	bufP = bufQ;
	if ((bufP < (nw_buffer_t)0x400000) || (bufP > (nw_buffer_t)0x7fffffff)) {
	  /* certain failure on MIPS */
	  fprintf(stderr, "WR.Invalid buf_next pointer: 0x%x, ignoring\n",
		  bufP);
	  bufP = 0;
	}
      } /* for */
      if (is_remote)
	(void) midway_uxpass_remote();
    }
    rc = nw_buffer_deallocate(nwbuf->peer.local_ep, nwbuf);
    if (rc != NW_SUCCESS) {
      fprintf(stderr,"nw_buffer_deallocate failed, rc %d (%s)\n",
	      rc, nw_result_char(rc));
      botch("nw_buffer_deallocate");
    }

#if FLOW_TIMES
    ft_rcv_time_rel = FT_READ_TIMER;
#endif

    /*
     * OK - try to process
     */
    done =  (filt(buf,len,cpti,(unsigned int)buf) == 0);

    /* DEBUG PROBLEM:  VERIFY THAT IF DONE, IS DONE! */
    if (done) {
      COMMMSG *cm = (COMMMSG *) (buf);
      int xcpti = cm->rid;
      extern cpt_t *cpt;
      int invals;

      if (xcpti != cpti) {	/* very interesting */
	char xxx[100];
	sprintf(xxx,"%d: done with unequal cptis [%d!=%d]",
		Comm_index, cpti, xcpti);
	botch(xxx);
      }
      invals = cpt[cpti].acksRequired[0] ^ cpt[cpti].acksReceived[0]; /*XXX*/
      if (invals != 0) {	/* interesting case */
	midClearBit((&invals), Comm_index); /* remove self */
	if (invals) {		/* still a problem */
	  char xxx[100];
	  sprintf(xxx,"%d: filter says done, but invs required 0x%08x= rcv !0x%08x", Comm_index, cpt[cpti].acksRequired[0], cpt[cpti].acksReceived[0]);
	  botch(xxx);
	}
      }
    } /* test done case  END DEBUG */

  } /* while not all packets received */
  mid_pbuf_free(pb);

  mutex_lock(midway_atm_fastlock);
  midway_atm_wantreply_active--; /* not here anymore */
  if (midway_atm_wantreply_queued)
    condition_signal(midway_atm_fastcond); /* let another waiter go */
  mutex_unlock(midway_atm_fastlock);

  return(1);
} /* midway_atm_wantreply */


#if 0   /* THIS IS NO LONGER USED! HISTORICAL READING ONLY */
/*
 * Plan C: for replies, use special call rather than queueing on cpt array.
 *	we take over atm_loop functions, call getmsg, and hope the reply
 *	comes in (by supplying a function for atm_upcall).  If all
 *	succeeds, we don't context switch.  If we fail, well we've just
 *	done a little extra work before blocking.
 *
 *      problem: mabye succeeded between time decided to get here, 
 *	and grabbing getlock?  could be...
 */
midway_atm_set_filter(comm_filter filt, any_t arg)
{
#if ATM_USE_INTERRUPTS_IN_WAIT == 0
  MIDWAY_ATM_BLOCK_INTERRUPTS;
#endif
  if (mutex_try_lock(midway_atm_getlock)) {
    if (midway_atm_filter == 0) {
      midway_atm_filter = filt;
      midway_atm_filter_arg = arg;
      midway_atm_filter_rc = 0;
      midway_atm_filter_used = 0;
    }
#if ATM_USE_INTERRUPTS_IN_WAIT == 0
    else {
      MIDWAY_ATM_ALLOW_INTERRUPTS;
    }
#endif
    mutex_unlock(midway_atm_getlock);
  }
#if ATM_USE_INTERRUPTS_IN_WAIT == 0
  else {
    MIDWAY_ATM_ALLOW_INTERRUPTS;
  }
#endif
} /* midway_atm_set_filter*/
midway_atm_clear_filter(comm_filter filt, any_t arg)
{
  int succ = 0;
  mutex_lock(midway_atm_getlock);
  if ((filt == midway_atm_filter) && (arg == midway_atm_filter_arg)) {
    midway_atm_filter = 0;
    midway_atm_filter_arg = 0;
    succ = 1;
  }
  mutex_unlock(midway_atm_getlock);
#if ATM_USE_INTERRUPTS_IN_WAIT == 0
  if (succ) {
    MIDWAY_ATM_ALLOW_INTERRUPTS;
  }
#endif

} /* midway_atm_clear_filter*/
midway_atm_wantreply(comm_filter filt, any_t arg)
{
  int spins;
  int queued;
  int onq;
  int succ = 0;
  extern task_t midway_task_self;	/* mach_task_self() */

  if (midway_atm_filter) {	/* already one installed */
    if ((midway_atm_filter != filt) && 
	(midway_atm_filter_arg != arg))
      return 0;			/* go slow */
    else
      succ = 1;
  }

#if ATM_USE_INTERRUPTS_IN_WAIT == 0
  if (!succ) {
#endif
    MIDWAY_ATM_BLOCK_INTERRUPTS;
#if ATM_USE_INTERRUPTS_IN_WAIT == 0
  }
#endif

  if (mutex_try_lock(midway_atm_getlock) == 0) {
    MIDWAY_ATM_ALLOW_INTERRUPTS; /* best to be safe here */
    midway_atm_filter_mumiss++;
    /*task_suspend(midway_task_self);*/ /* real quick debug */
    return 0;			/* failed to get mutex */
  }

  /* should do this test only after interrupts turned off. */
  if (midway_atm_filter) {	/* already set */
    if (midway_atm_filter_used && (midway_atm_filter_rc != 1/*do more*/)) {
      int rc = midway_atm_filter_rc;
      midway_atm_filter = 0;
      midway_atm_filter_arg = 0;
      mutex_unlock(midway_atm_getlock);
      MIDWAY_ATM_ALLOW_INTERRUPTS;
      return(rc);
    }
  } else {
    midway_atm_filter = filt;
    midway_atm_filter_arg = arg;
    midway_atm_filter_used = 0;
  }

  midway_atm_locker = 2;
  midway_atm_filter_muhit++;
  midway_atm_upcall_queued = 0;


#define OURSPIN (int)0x6fffffff /* 10000, 100000, ... */
  spins = /*midway_atm_loop_spins*/ OURSPIN;

  while(spins > 0) {
    midway_atm_nopack = 1;	/* assume no packets */
    midway_atm_filter_rc = 0;	/* assume didn't work */
    get_messages();		/* do it */
    if (midway_atm_filter_rc == 0) {
      if (midway_atm_nopack == 0) {
	/* got one that wasn't one we were looking for */
	break;
      }
      else {
	/* didn't get one */
	spins--;
      }
    }
    else if (midway_atm_filter_rc == -1) {
      /* filter worked, and all messages have arrived */
      int spn = OURSPIN - spins;
      if (spn < midway_atm_filter_minspin)
	midway_atm_filter_minspin = spn;
      if (spn > midway_atm_filter_maxspin)
	midway_atm_filter_maxspin = spn;
      break;
    }
    else {			/* not 0 or -1, must be +1 */
      spins = OURSPIN;		/* reset spins */
    }
  } /* while */
  
  midway_atm_filter = (comm_filter) 0; /* clear filter */

  queued = midway_atm_upcall_queued;
  midway_atm_upcall_queued = 0;
  if (queued == 0)
    onq = (MID_PBUF_QTOP(midway_atm_bufq) != 0);
  else
    onq = 1;
  mutex_unlock(midway_atm_getlock);
  MIDWAY_ATM_ALLOW_INTERRUPTS;

  /* do exact number of signals instead? */
  if (midway_atm_intr_waiting == 0) { /* "normal" case */
    if (onq)
      condition_signal(midway_atm_c);
  } else {			/* probable outstanding request */
    /* following # could be "low" */
    midway_atm_intr_waitsig += (queued - 1); /* # "extra" signals */
    while (queued) {
      condition_signal(midway_atm_c);
      queued--;
    }
  }

  return(midway_atm_filter_rc != 0);
} /* midway_atm_wantreply */



/*
 * Plan E: Get interrupts working by hook or by crook.  New interface:
 * thread that gets interrupt.  We will turn off interrupts when
 * looking for a specific reply, though.
 */
kern_return_t
midway_atm_interrupt(int *arg,	/* passed in from atm_message_wait */
		     int reason /* reason for the interrupt */
		     )
{
  /* like midway_atm_poll, for now, but we always return 0 */
  int queued;
  int onq;
  extern task_t midway_task_self;	/* mach_task_self() */

#if oldway			/* we want to continue now */
  if (midway_atm_filter) {	/* one installed */
    midway_atm_intr_filter++;
    return 0;			/* go slow */
  }
#endif

  if (mutex_try_lock(midway_atm_getlock) == 0) {
    midway_atm_intr_mumiss++;
    /*task_suspend(midway_task_self);*/ /* real quick debug */
    return 0;			/* failed to get mutex */
  }

  midway_atm_locker = 4;
  midway_atm_intr_muhit++;
  midway_atm_upcall_queued = 0;

  midway_atm_nopack = 1;	/* assume no packets */
  get_messages();		/* do it */
  
  queued = midway_atm_upcall_queued;
  midway_atm_upcall_queued = 0;

  if (queued) {
    midway_atm_intr_succ++;
    midway_atm_intr_pkts += queued;
    onq = 1;
  }
  else {
    onq = (MID_PBUF_QTOP(midway_atm_bufq) != 0);
  }
  mutex_unlock(midway_atm_getlock);

  if (midway_atm_intr_waiting == 0) { /* "normal" case */
    if (onq)
      condition_signal(midway_atm_c);  /* since high prio, should not desched*/
  } else {			/* probable outstanding request */
    /* following # could be "low" */
    midway_atm_intr_waitsig += (queued - 1); /* # "extra" signals */
    while (queued) {
      condition_signal(midway_atm_c);
      queued--;
    }
  }

  return(KERN_SUCCESS);
  
} /*midway_atm_interrupt*/

/*
 * midway_atm_willblock(): we only signal a waiter, though there could be
 * multiple things queued.  The problem is that the waiter could block
 * (usually because of needing alloc info); in fact many waiters could
 * block.  To keep the ball rolling, we want to signal someone to handle
 * messages if we block.  Unfortunately, I don't know how to do this
 * exactly -- the xpager is single threaded, and many requests could
 * block waiting for info without our being notified of it.
 *
 * This is therefore a hint... whenever a CommRequest is going to block,
 * we are called.  Since CR is going to block, it will let someone else
 * through (we will signal once because wantreply is what was done first).
 * So, if CR is going to block, and there are many things on the queue,
 * signal to rest.  Also set bit so that future interrupts will signal
 * all.
 */
midway_atm_willblock()
{
  int n;

  if (midway_atm_rcv_active <= 0)
    return 0;			/* no need to do anything special. */

  mutex_lock(midway_atm_intr_wait_lock);
  midway_atm_intr_waiting++;
  midway_atm_intr_waitnum++;	/* note this for posterity */
  mutex_unlock(midway_atm_intr_wait_lock);

  n = midway_atm_queued - 1;
  if (n > 0) {
    for (; n; n--) {
      midway_atm_intr_waitsig++;	/* note # signals for posterity */
      condition_signal(midway_atm_c);
    }
  }
} /* midway_atm_willblock()*/


midway_atm_doneblock()
{
  mutex_lock(midway_atm_intr_wait_lock);
  midway_atm_intr_waiting--;
  mutex_unlock(midway_atm_intr_wait_lock);
}

#endif /* 0 - delete above section HISTORICAL */

mid_atm_print_stats()
{
  /* sick of adding stuff to comm.c.  Have comm call here. */
  char smsg[255];
  u_long atbats;

#if 0 /* old way */
  sprintf(smsg, "wantreply filter: %d 0x%x minspin %d 0x%x maxspin",
	  midway_atm_filter_minspin,
	  midway_atm_filter_minspin,
	  midway_atm_filter_maxspin,
	  midway_atm_filter_maxspin);
  StatusMsg(smsg);

  sprintf(smsg, "%d mumiss %d muhit %d vctx %d cspins %d wakegood %d spureous",
	  midway_atm_filter_mumiss,
	  midway_atm_filter_muhit,
	  midway_atm_vol_ctxsw,
	  midway_atm_ctx_spins,
	  midway_atm_rcv_wakeups,
	  midway_atm_rcv_spureous);
  StatusMsg(smsg);

  atbats = midway_atm_poll_filter 
    + midway_atm_poll_mumiss + midway_atm_poll_muhit;
  sprintf(smsg, 
    "atm_poll: %d fltr %d mumiss %d muhit %d data %d pkts (%4.2f%% hit rate)",
	  midway_atm_poll_filter, midway_atm_poll_mumiss, 
	  midway_atm_poll_muhit, midway_atm_poll_succ, midway_atm_poll_pkts,
	  (atbats == 0) ? 0.0 :
	  (double)((((double)midway_atm_poll_succ)*100.0)/(double)atbats));
  StatusMsg(smsg);

  atbats = midway_atm_intr_filter 
    + midway_atm_intr_mumiss + midway_atm_intr_muhit;
  sprintf(smsg, 
    "atm_intr: %d fltr %d mumiss %d muhit %d data %d pkts (%4.2f%% hit rate)",
	  midway_atm_intr_filter, midway_atm_intr_mumiss, 
	  midway_atm_intr_muhit, midway_atm_intr_succ, midway_atm_intr_pkts,
	  (atbats == 0) ? 0.0 :
	  (double)((((double)midway_atm_poll_succ)*100.0)/(double)atbats));
  StatusMsg(smsg);
#endif /*0*/

  sprintf(smsg,"atm_snd: %d overruns, max spins %d avg spins %d",
	  midway_atm_snd_overruns, midway_atm_snd_maxspins, 
	  (midway_atm_snd_overruns?
	   midway_atm_snd_totspins/midway_atm_snd_overruns:0));
  StatusMsg(smsg);

  sprintf(smsg,"atm_snd: %d peer %d rpc-reply %d metadata",
	  midway_atm_snd_peer, midway_atm_snd_rpc,
	  midway_atm_snd_meta);
  StatusMsg(smsg);

  sprintf(smsg,"atm_wantreply: %d received %d punted",
	  midway_atm_rcv_rpc, midway_atm_rcv_rpc_punt);
  StatusMsg(smsg);
} /* mid_atm_print_stats */

/* code copied from comm.c.  Copied, so we don't have to release this
   function to the world. */
/*
 ************************************************************************
 *		m i d w a y _ s e t _ g o o d _ p r i o r i t y		*
 *----------------------------------------------------------------------*
 * I would like this to be in a separate file, but by placing it here   *
 * I don't have to worry about it being a global symbol.  		*
 *									*
 * Reset max priority of this task downward (range 0-31, smaller better)*
 * and this thread in particular to a very nice value.			*
 *									*
 * This task must be running in group kmem or uid 0 to use task_by_pid. *
 *									*
 ************************************************************************
 */
static int			prio_init = 0;	/* initialized? */
static mach_port_t 		host_priv;	/* priv host port */
#if MIDWAY_WIRE_MEMORY
mach_port_t			midway_host_priv; /* make it global, sigh. */
#endif
static processor_set_name_t 	assigned_set;	/* proc set we are asgned to */
static processor_set_t		proc_set;	/* ctl port for our proc set */
static struct mutex prio_mutex = MUTEX_INITIALIZER; /* cheat! */
static mutex_t pmu = &prio_mutex;

int midway_atm_rcv_prio = 1;	/* global for now, leaving 1 as "maximum". */
int midway_maxpriority = 1;	/* global for now, leaving 1 as "maximum". */
int midway_fixpri_quantum = 100; /* mSec */
int midway_fixpri_enable = 1;	/* allow to occur */
int midway_fixpri_enable_all = 1; /* enable for "all" */ /* was 0 */

#define FAILONERR(s)	if (krc != KERN_SUCCESS) { \
				mach_error(s, krc); \
				return(1);}
#define MFAILONERR(s)	if (krc != KERN_SUCCESS) { \
				prio_init = -1; \
				mach_error(s,krc); \
				mutex_unlock(pmu); \
				return(1);}

static int
midway_init_priority()
{
  kern_return_t krc;

  if (prio_init == 0) {
    mutex_lock(pmu);
    if (prio_init == 0) {	/* really not initialized */
      host_priv = task_by_pid(-1); /* back door to get priv host port */
      if (host_priv < 0) {	/* Brian did this.  Not sure if proper. */
	perror("bsd_task_by_pid");
	prio_init = -1;		/* show error */
	mutex_unlock(pmu);
	return(1);
      }
      if (host_priv == 0) {	/* see this empirically */
	StatusMsg("Can't get priv host port [not grp=kmem or uid=0?], will not set receive thread priority.");
	prio_init = -1;
	mutex_unlock(pmu);
	return(1);
      }

#if MIDWAY_WIRE_MEMORY
      midway_host_priv = host_priv;
#endif

      krc = task_get_assignment(mach_task_self(), &assigned_set);
      MFAILONERR("task_get_assignment");

      krc = host_processor_set_priv(host_priv, assigned_set, &proc_set);
      MFAILONERR("host_processor_set_priv");

      if (midway_fixpri_enable) {
	krc = processor_set_policy_enable(proc_set, POLICY_FIXEDPRI);
	MFAILONERR("processor_set_policy_enable");
      }

      prio_init = 1;		/* Done, success */
    }
    mutex_unlock(pmu);
  } /* not initialized */
  return(0);
} /* midway_init_priority */
int
midway_set_good_priority(int pri)
{
  kern_return_t krc;

  midway_init_priority();

  if (prio_init < 0)
    return(1);			/* failure */

  /* make fixed if necessary */
  if (midway_fixpri_enable){
    krc = thread_policy(mach_thread_self(), POLICY_FIXEDPRI, 
			midway_fixpri_quantum);
    if (krc == KERN_FAILURE) {
      StatusMsg("warning: no POLICY_FIXEDPRI");
    } else {
      MFAILONERR("thread_policy");
    }
  }

  /* set max priority down  */
  krc = thread_max_priority(mach_thread_self(), proc_set, 
			    midway_maxpriority);
  FAILONERR("thread_max_prio");
  
  /* now we can make our priority better */
  krc = thread_priority(mach_thread_self(), pri, FALSE);
  FAILONERR("thread_priority");

  return(0);			/* success */
} /* midway_set_good_priority */
int
midway_set_good_priority_all(int pri)
{
  kern_return_t krc;
  thread_array_t thread_list;
  mach_msg_type_number_t thread_count;
  int i;

  midway_init_priority();

  if (prio_init < 0)
    return(1);			/* failure */

  krc = task_threads(mach_task_self(), &thread_list, &thread_count);
  FAILONERR("task_threads");

  for (i=0; i<thread_count; i++) {

    /* make fixed if necessary */
    if (midway_fixpri_enable && midway_fixpri_enable_all){
      krc = thread_policy(thread_list[i], POLICY_FIXEDPRI, 
			  midway_fixpri_quantum);
      if (krc == KERN_FAILURE) {
	StatusMsg("warning: no POLICY_FIXEDPRI");
      } else {
	MFAILONERR("thread_policy");
      }
    }

    /* set max priority down  */
    krc = thread_max_priority(thread_list[i], proc_set, 
			      midway_maxpriority);
    FAILONERR("thread_max_prio");
  
    /* now we can make our priority better */
    krc = thread_priority(thread_list[i], pri, FALSE);
    FAILONERR("thread_priority");
  }

  VM_DEALLOCATE(mach_task_self(), (vm_address_t) thread_list,
		thread_count*sizeof(thread_t));

  return(0);			/* success */
} /* midway_set_good_priority */
