/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * try to trace the packet processing time through the library.
 * Use the frc.
 * 4 dec 92, mjz
 */

#define FLOW_TIMES 0

extern u_int *midway_kern_frc;

/* 5000/200 only, is bus cycle time */
#define FT_FRC_TICKS_PER_USEC 25 /* 40nS */

#if FLOW_TIMES

#define FT_TYPE  0
#define FT_LOCK  1
#define FT_TID	 2

#define FT_START 3
#define FT_ATMRECV		3
#define FT_ATMCOPY		4
#define FT_ATMREL		5
#define FT_COMMRECV             6
#define FT_WORKER 		7

/* requests */
#define FT_LOCKREQ		8
#define FT_LOCKXFER		9
#define FT_LOCKQRS		10
#define FT_LOCKDLTA		11
#define FT_LOCKDLOOP		12
#define FT_LOCKXFERDONE		13
#define FT_WORKERUPDT		14
#define FT_FREE			15

#define FT_REQPNUM		16 /* # of req packets */
#define FT_REQPAKS		17 /* sequence of start, aftercopy pairs */

/* replies */
#define FT_SENDSTART		8  /* time at which request is started */
#define FT_SENDREQ		9  /* just before send (so overhead) */
#define FT_SENDALLOC		10 /* allocate buffer time */
#define FT_SENDCOPY		11 /* copy to network buffer (+ spin) */
#define FT_SENDSEND		12 /* after nw_send completes */
#define FT_SENDDONE		13 /* time after send (good for filters) */
#define FT_REPSIGNAL		14 /* when q'd guy is signalled */
#define FT_REQDONE		15 /* when request is done */

#define FT_REPPNUM		16 /* # of packets in list */
#define FT_REPPAKS		17 /* packets for reply - start, done pairs */


#define FT_MAX   804
#define FT_MAXPAKS 131		/* floor((FT-MAX - 18)/6) */
#define FT_NUM	100		/* keep 100 packets worth */

extern u_int ft_times[FT_NUM][FT_MAX];
extern u_int ft_next;		/* next index into ft_times */
extern mutex_t ft_mu;
extern unsigned int ft_send_time_start; /* in atm_iface */
extern unsigned int ft_send_time_end;

#define DEFINE_FT u_int ft_times[FT_NUM][FT_MAX] = { 0 }; \
                  u_int ft_next = 1; \
                  mutex_t ft_mu

#define INITIALIZE_FT { midway_setup_mapped_frc(); \
			  ft_next = 0; \
			    ft_mu = mutex_alloc();\
			      mutex_set_name(ft_mu,"ft_mu"); }

#define NEXT_FT {mutex_lock(ft_mu); \
		   fti = (ft_next < FT_NUM)?ft_next++:0; mutex_unlock(ft_mu);}
#define FT_READ_TIMER *midway_kern_frc
#define FT_MSGINFO(t,i) &(ft_times[fti][(t)+(6*(i))])

/* following assumes we only wrap once */
#define FT_DIFF(x,y) (((x)==0)?0:(((y)==0)?0:(((x)>=(y))?x-y:x+(0xffffffff-(y)) ) ) )
#endif /* FLOW_TIMES */
/* need the following for lock call (ugh) */
#define FT_STASH_INMSG version	/* use version field of msgheader struct */
