/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
#include <mach.h>
#include <sys/types.h>
#include <stdio.h>

#define D_SUCCESS 0

static int mapped_init = 0;

unsigned int midway_kern_frc_dummy = 0;
unsigned int *midway_kern_frc = &midway_kern_frc_dummy;


kern_return_t
midway_setup_mapped_frc() {
  vm_offset_t *counter_p;
  kern_return_t   rc;
  mach_port_t     device_server_port, device_port, pager;
  vm_machine_attribute_val_t cacheit;


  if (mapped_init)
    return KERN_SUCCESS;


/*  printf("FRCounter initialization\n"); */
  device_server_port = task_by_pid(-2);
  if (device_server_port == MACH_PORT_NULL) {
    fprintf(stderr,"Error getting device port-- group kmem or root?\n");
    return -1;
  }

  rc = device_open(device_server_port, 0, "frc", &device_port);
  if (rc != D_SUCCESS) {
    mach_error("Error opening frc",rc);
    return rc;
  }

  pager = MACH_PORT_NULL;
  rc = device_map(device_port, VM_PROT_READ,
                  0, vm_page_size, &pager, 0);
  if (rc != D_SUCCESS) {
    mach_error("Error mapping frc",rc);
    return rc;
  }

  counter_p = (vm_offset_t *) &midway_kern_frc;
  rc = vm_map(mach_task_self(),
              counter_p, vm_page_size, 0, TRUE,
              pager, 0, FALSE,
              VM_PROT_READ, VM_PROT_READ, VM_INHERIT_NONE);
  if (rc != D_SUCCESS) {
    mach_error("Error vm_map frc",rc);
    return rc;
  }

  rc = mach_port_deallocate(mach_task_self(), pager);
  if (rc != KERN_SUCCESS) {
    mach_error("Error deallocating frc pager",rc);
    return rc;
  }

  /* Now make it uncached */
  cacheit = MATTR_VAL_OFF;
  rc = vm_machine_attribute(mach_task_self(), *counter_p, vm_page_size,
                            MATTR_CACHE, &cacheit);
  if (rc != KERN_SUCCESS) {
    mach_error("Error uncaching frc",rc);
  }
  midway_kern_frc = (unsigned int *) *counter_p;
  mapped_init = 1;

  return rc;
} /* midway_setup_mapped_frc */

  
    
  
