/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * maptime.c: utility to set up a mapped gettimofday like timer.
 *
 *  9 Nov 92 (mjz): steal from my more general 'timing.c'
 */
/*#include "timing.h" */
#ifndef KERNEL
#define MY_KERNEL_YAYA 1
#define KERNEL 1		/* mach 3 bogosity */
#endif
#include <sys/time.h>		/* for timevals, and other good things */
#ifdef MY_KERNEL_YAYA
#undef KERNEL
#endif
typedef struct timeval tv_t;
#include <stdio.h>
#include <mach.h>
#include <mach/message.h>
#include <mach/error.h>
#include <device/device_types.h>
#include <mach/time_value.h>

static mapped_init = 0;
volatile struct mapped_time_value *mtime;

midway_setup_mapped_timer()
{
    long secs, usecs;
    kern_return_t rc;
    mach_port_t device_server_port, device_port, pager = MACH_PORT_NULL;

    /* gak backdoor.  need kmem group/root to doo this. */
    device_server_port = task_by_pid(-2);
    rc = device_open(device_server_port,D_READ,"time",&device_port);
    if (rc != KERN_SUCCESS) {
/*      printf("line:%d\n", __LINE__);*/
      printf("device_open error: %s\n", mach_error_string(rc));
      exit(1);
    }

    rc = device_map(device_port, VM_PROT_READ,
		    0, vm_page_size, &pager, 0);
    if (rc != D_SUCCESS) {
/*      printf("line:%d\n", __LINE__);*/
      printf("device_map error: %s\n", mach_error_string(rc));
      exit(1);
    }
    if (pager == MACH_PORT_NULL) {
/*      printf("line:%d\n", __LINE__); */
      printf("device_map (no pager) error: %s\n", mach_error_string(rc));
      exit(1);
    }


    rc = vm_map(mach_task_self(),
		(vm_offset_t *) &mtime, sizeof *mtime, 0, TRUE,
		pager, 0, FALSE,
		VM_PROT_READ, VM_PROT_READ, VM_INHERIT_NONE);
    if (rc != D_SUCCESS) {
/*      printf("line:%d\n", __LINE__);*/
      printf("vm_map error: %s\n", mach_error_string(rc));
      exit(1);
    }

    rc = mach_port_deallocate(mach_task_self(), pager);
    if (rc != KERN_SUCCESS) {
/*      printf("line:%d\n", __LINE__);*/
      printf("mach_port_deallocate error: %s\n", mach_error_string(rc));
      exit(1);
    }

    mapped_init = 1;

} /* setup_mapped_timer */

midway_gettimeofday(tv_t *tp, struct timezone* tzp)
{
  if (!mapped_init) {
    midway_setup_mapped_timer();
  }

  do {				/* see mach/time_value.h */
    tp->tv_sec = mtime->seconds;
    tp->tv_usec = mtime->microseconds;
  } while (tp->tv_sec != mtime->check_seconds);
  /* we don't do timezones. */
}/* mapped_gettimeofday */ 
