/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * midway_comp.h:  Definitions shared between midway compiler and runtime
 *
 * HISTORY
 * 20 Sep 93 (was): Initial version.
 */

#if !defined(__MIDWAY_COMP_H__)
#define __MIDWAY_COMP_H__

#undef MIDWAY_DEBUG				/* define to enable midway debugging in compiler */

#define MIDWAY_REGION_SHIFT                  23	          /* 1<<23 == 8 MB */
#define MIDWAY_REGION_SIZE   (1 << MIDWAY_REGION_SHIFT)   /* size of midway shared data region */

#define MIDWAY_TEMPLATE_SIZE               4096    /* size of dirty bit template = 1 page */

#define MIDWAY_TEMPLATE_WORD_OFFSET         512    /* offset to builtin_word w/in template */
#define MIDWAY_TEMPLATE_DOUBLE_OFFSET         0    /*           builtin_double */
#define MIDWAY_TEMPLATE_HALF_OFFSET        1024    /*           builtin_half   */
#define MIDWAY_TEMPLATE_BYTE_OFFSET        1536    /*           builtin_byte   */
#define MIDWAY_TEMPLATE_AREA_OFFSET        2048    /*           builtin_area   */

#define MIDWAY_TEMPLATE_VALUE "MidwayStaticDBit" /* initial value for static template. MUST BE 16 BYTES LONG */
#define MIDWAY_TEMPLATE_ZERO_SIZE  (MIDWAY_TEMPLATE_SIZE - 16)

#define MIDWAY_TEMPLATE_VALID_P(x)	\
  (strcmp(((char *)x),MIDWAY_TEMPLATE_VALUE) == 0)



#ifndef ASSEMBLER
typedef char midway_template_t[MIDWAY_TEMPLATE_SIZE];

extern unsigned long midway_region_size; 	/* sizeof midway region (default = MIDAY_REGION_SIZE)*/
extern unsigned long midway_region_mask; 	/* = ~(midway_region_size-1) */
#endif

#endif /* __MIDWAY_COMP_H__ */





