/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * midway_mem.h: things related to memory allocation that are not really

 * 	"private" to shmalloc (those are in shmalloc.h)
 *
 * Originally by bnb, sometime in june
 *
 *  8 Aug 91 (mjz): add midway_default_allocation_size
 * 22 Aug 91 (mjz): renamed midway_mem from midway, for merge of runtime code
 *  2 Aug 92 (mjz): change default static linesize to 4 from 16
 * 11 Aug 92 (was): added clock_t type
 *  9 Nov 92 (was): minor correction for using larger regions
 * 24 Feb 93 (was): lazy allocate all regions to processors
 */
#ifdef OMITSHARED
#define shared
#endif

shared char *shmalloc();

clock_t midway_get_local_clock();
clock_t midway_set_local_clock();



int midway_cache_line_size();	        /* return how big lines are for the specified address */
int midway_line_dirty();		/* return if line is dirty */
void midway_clean_line();		/* mark line clean */
clock_t *midway_dirty_info_address();	/* return address of line information */

/*
 * Programs can modify the static linesize prior to running.
 */
#define MIDWAY_DEFAULT_STATIC_LINESIZE 4
extern int midway_default_static_linesize;

/*
 * Define number of regions to pre-allocate at each processor.
 */
#define MIDWAY_DEFAULT_ALLOCATION_REGIONS   0  /* # of regions per processor */
extern unsigned long midway_default_allocation_size;




