/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * packet buffer header file...
 *
 * 18 mar 91 (mjz): created out of pbuf.c (which has been around for awhile)
 * 16 jul 91 (mjz): put pbufe_t here, added queue stuff (for barriers)
 * 15 Oct 92 (was): changed to determine pbuf size from network max.
 */
#ifndef _PBUF_H_
#define _PBUF_H_ 1

#include "comm.h"		/* for COMMMSG definition */

/*#define PBUF_VERIFY 1	*/	/* define to compile in pbuf addr checks */



/*
 * Define max message & data sizes
 */
#define NETWORK_MSG_OVERHEAD (20+8) /* communication overhead per message
		 		     * sizeof(ip_header) + sizeof(udp_header)*/
#define MIDWAY_MSG_OVERHEAD sizeof(COMMMSG)

#define MAX_PBUF_SIZE ((8 * 1024) - NETWORK_MSG_OVERHEAD)
#define MAX_DATA_SIZE (MAX_PBUF_SIZE - MIDWAY_MSG_OVERHEAD)

extern int mid_max_data_size;       /* Max data allowed in buffer */
extern int mid_max_pbuf_size;       /* actual size of buffer */


typedef char *pbuf;		/* really, char pbuf[MAX_DATA_SIZE] */

/* internal - what a pbuf really is */
typedef struct pbufe {
  struct pbufe *next;
  unsigned long	sig;		/* keep on x8 boundry too */
  char data[4];                 /* data buffer, actual size is pbuf_size */
} pbufe_t;


typedef struct pbufq {
  pbufe_t *head;		/* first one */
  pbufe_t **last;		/* where next one will be linked in */
} pbufq_t, *pbuf_q;

extern int pbuf_init();		/* initialize packetbuffer pool */
extern pbuf mid_pbuf_get();	/* get a packet buffer */
extern int mid_pbuf_free(/*pbuf x*/);	/* release a packet buffer */

extern void mid_pbuf_qinit(/*pbufq q*/); /* init a pbufq_t */
extern void mid_pbuf_enqueue(/*pbuf p, pbufq q*/); /* place p on q */
extern pbuf mid_pbuf_dequeue(/*pbuf_q q*/); /* grab head of q */
extern void mid_pbuf_qfree(/*pbufq q*/); /* free all entries; re-init... */
extern void mid_pbuf_qmove(/*pbufq q1,q2*/); /* move q1 to q2 */

/* following for running down Q and doing something (like sending off) */
extern pbufe_t *mid_pbuf_qtop(/*pbuf_q q*/); /* get top, return entry, !pbuf */
extern pbufe_t *mid_pbuf_qnext(/*pbufe_t e*/); /* get next one */
extern pbuf mid_pbuf_qtail(/*pbuf_q q*/); /* get last pbuf */

#define Q2PBUF(x) ((x)->data)
/* macro versions of some Q manipulations */
#define MID_PBUF_QTOP(q) ((q)->head)
#define MID_PBUF_QNEXT(e) ((e)->next)
#define MID_PBUF_QTAIL(q) (((pbufe_t *)((q)->last))->data)


#ifdef PBUF_VERIFY
#  define ASSERT_PBUF_ADDR(addr,len) mid_pbuf_verify(addr,len);
#else
#  define ASSERT_PBUF_ADDR(addr,len)
#endif

#endif _PBUF_H_

