/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1989 Carnegie-Mellon University
 * All rights reserved.  The CMU software License Agreement specifies
 * the terms and conditions for use and redistribution.
 */
/*
 * HISTORY
 * $Log: mcount.s,v $
# Revision 1.2  1992/11/13  17:35:39  mattz
# Update for gcc bug (?): when compiled -O -pg, CommSend loads two
# important values into v0 and v1, and then calls _mcount.  _mcount
# eventually calls mcountaux, which calls mutex_try_lock, which trashes
# v0 ('cause it is a return register).  I believe v1 is trashed too.
#
# For now, save v0 and v1 in _mcount.
#
# Revision 1.1  1992/10/08  19:57:15  wsawdon
# Added support for gprof.
# New routine mcount, to count procedure entries.
#
 * Revision 2.3.2.1  92/05/03  22:19:34  cmaeda
 * 	Added _mcount function.
 * 
 * Revision 2.3.1.1  92/04/14  13:52:31  cmaeda
 * 	Added _mcount function.
 * 
 * Revision 2.3  90/03/14  21:28:07  rwd
 * 	Added INCLUDE_VERSION.
 * 	[90/01/24            af]
 * 
 * Revision 2.2  89/11/29  15:28:45  af
 * 	Created.
 * 	[89/11/06            af]
 * 
 */

/*
 *	Miscellaneous assembly code.
 */

#include <mips/asm.h>


/*
 *	Object:
 *		INCLUDE_VERSION			EXPORTED absolute
 *
 *  Arrange for the include file version number to appear directly in
 *  the name list.
 *
 */
#include <sys/version.h>
	ABS(_INCLUDE_VERSION, INCLUDE_VERSION)
#undef  INCLUDE_VERSION
        ABS(INCLUDE_VERSION,_INCLUDE_VERSION)


/*
 * 	Object:
 *	 _mcount			EXPORTED absolute
 *
 *	This functions should never be called from C.
 *	It constructs a C-style arguments list from registers
 * 	and calls the C function, mcountaux.
 *
 * 	XXX Calling C from assembly is probably a bad way to do things.
 *      XXX C makes some assumptions about register usage which require
 *	XXX us to save a lot of stuff on the stack.
 *
 *	This function also assumes it is invoked by the following
 *	code sequence:
 *
 *		move	at,ra
 *		jal	_mcount
 *		subu	sp,8		# _mcount pops 2 arguments
 *
 */
/* BAD_COMPILER: current gcc on mips loads v0 & v1, then calls mcount if
 * we are optimizing.  bah. 10nov92mjz
 */
#define BAD_COMPILER 1
	.align	2
	.globl	_mcount
	.ent	_mcount
	.extern gprof_do_profiling 4
	.set	noat
	.set	noreorder
_mcount:
#if BAD_COMPILER
	.mask (M_RA|M_A3|M_A2|M_A1|M_A0|M_V1|M_V0|M_AT),-4
	.frame	sp,40,ra	# 8 saved regs, 2 arguments
#else
	.mask (M_RA|M_A3|M_A2|M_A1|M_A0|M_AT),-4
	.frame	sp,32,ra	# 6 saved regs, 2 arguments
#endif
	/*
	 * Try to get out fast if profiling is disabled.
	 */
	lw	t0,gprof_do_profiling	# should go through gp
	nop
	bne	t0,zero,1f	
	addu	sp,8
	j	ra
	move	ra,AT	

1:
#if BAD_COMPILER
	subu	sp,40		# caller already made room for 2
	sw	ra,36(sp)	# frameoffset is 36
#else
	subu	sp,32		# caller already made room for 2
	sw	ra,28(sp)	# frameoffset is 28
#endif
	/*
	 * Since we're about to call a C function,
	 * we have to save any registers that the
	 * C compiler assumes it can just trash.
	 * Ask me sometime how many days it took to figure this out.
	 */
#if BAD_COMPILER
	sw	a3,32(sp)
	sw	a2,28(sp)
	sw	a1,24(sp)
	sw	a0,20(sp)
	sw	v1,16(sp)
	sw	v0,12(sp)
	sw	AT,8(sp)
#else
	sw	a3,24(sp)
	sw	a2,20(sp)
	sw	a1,16(sp)
	sw	a0,12(sp)
	sw	AT,8(sp)
#endif
	
	move	a1,ra		# TO argument (caller's address)
	jal	mcountaux
	move	a0,AT		# FROM argument (caller's ra)

#if BAD_COMPILER	
	
	lw	AT,8(sp)
	lw	v0,12(sp)
	lw	v1,16(sp)
	lw	a0,20(sp)
	lw	a1,24(sp)
	lw	a2,28(sp)
	lw	a3,32(sp)
	lw	ra,36(sp)
	addu	sp,40		# pop 10 words
#else
	lw	AT,8(sp)
	lw	a0,12(sp)
	lw	a1,16(sp)
	lw	a2,20(sp)
	lw	a3,24(sp)
	lw	ra,28(sp)
	addu	sp,32		# pop 8 words
#endif
	j	ra
	move	ra,AT
	.end	_mcount
