/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * midway_end.s
 *
 * This file contains the system dependent code for finding the end of 
 * the application's static data. This file must be the LAST object 
 * file included by all midway applications.
 *
 *  22Feb94 (was): created.
 */

#include <mips/asm.h>

#define PAGE_BASED 1	/* standard gcc doesn't pass command line -D's to asm */

#if defined(PAGE_BASED)

/*
 * Define variables at the END of each of the data regions
 * See comments in midway_start.s for more info.
 */
#define MIPS_TAG_SIZE  16	/* MUST match _end.s & vm_paged.c */
#define MIPS_STAG_SIZE  4

/* define end of text region */
	.text
	.align 2
	.ent	midway_end_text_data,0
midway_end_text_data:
	.frame	sp,0,ra
	j	ra
	.end 	midway_end_text_data

/* define end of bss region */
	.globl	midway_end_bss_data
	.lcomm	midway_end_bss_data, MIPS_TAG_SIZE

/* define end of sbss region */
	.globl	midway_end_sbss_data
	.lcomm	midway_end_sbss_data, MIPS_STAG_SIZE

/* define variables in static regions to get pointers above */
	.lcomm	midway_end_static_bss_data, MIPS_TAG_SIZE
	.lcomm	midway_end_static_sbss_data, MIPS_STAG_SIZE

/* define variables in common regions */
	.comm	midway_end_comm_data, MIPS_TAG_SIZE
	.comm	midway_end_scomm_data, MIPS_STAG_SIZE

/* define end of data region */
	.data
	.align	2
	.globl	midway_end_data_data
midway_end_data_data:
	.space	MIPS_TAG_SIZE

/* define end of small data region */
	.sdata
	.align	2
	.globl	midway_end_sdata_data
midway_end_sdata_data:
	.space	MIPS_STAG_SIZE

/* 
 * Declare global pointers to end of each region
 * (starting address of boundary tag)
 */
	.globl	midway_end_text
midway_end_text:
	.word	midway_end_text_data

	.globl	midway_end_data
midway_end_data:
	.word	midway_end_data_data

	.globl	midway_end_sdata
midway_end_sdata:
	.word 	midway_end_sdata_data

	.globl midway_end_bss
midway_end_bss:
	.word	midway_end_bss_data

	.globl midway_end_sbss
midway_end_sbss:
	.word	midway_end_sbss_data

	.globl midway_end_static_bss
midway_end_static_bss:
	.word	midway_end_static_bss_data

	.globl midway_end_static_sbss
midway_end_static_sbss:
	.word	midway_end_static_sbss_data

	.globl	midway_end_comm
midway_end_comm:
	.word	midway_end_comm_data

	.globl	midway_end_scomm
midway_end_scomm:
	.word	midway_end_scomm_data


#endif /* PAGE_BASED */



