/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * stack.s
 *
 * This file contains the system dependent code for manipulating a thread's stack.
 * Since the compiler may inadvertently generate a dirtybit update call for 
 * data on the stack, the application's stack must have a dirtybit template
 * stored at its base. This is accomplished by moving the application's thread's
 * stack to a special region, which has a PRIVATE template stored at its base.
 *
 *
 * Stack on a MIPS machine grows down. For example:
 *
 *   Base = 0x480000   +-------------+
 *                     | / / / / / / |
 *                     | / U S E D / |
 *                     | / / / / / / |
 *   sp   = 0x47f100   +-------------+
 *                     |             |
 *                     |             |
 *                     |   F R E E   |
 *                     |             |
 *                     |             |
 *   End  = 0x400000   +-------------+
 *
 *  cthread_stack_size = 0x80000
 *  cthread_stack_mask = 0x7ffff
 *
 *  Stack Base = (sp | mask) + 1
 *
 *
 *  11Nov93 (was): created.
 */

#include <mips/asm.h>

/*
 * Declare external variables
 */
IMPORT(cthread_stack_mask, 4)
IMPORT(cthread_stack_size, 4)
IMPORT(vm_private_stackP, 4)
IMPORT(memcpy, 4)

	.text
	.align	2
	.set noreorder

/************************************************************************
*									*
*	Move a thread's stack. 						*
*									*
* vm_offset_t imove_stack 						*
* 	(								*
*		vm_offset_t stack_base					*
*	);								*
*									*
************************************************************************/

NESTED(imove_stack,16,ra)
	.mask (M_RA),-4
	addi	sp, -20			# 2 saved regs, 3 arguments
	sw	ra, 16(sp)

	#
	# On entry: a0 contains ptr to thread's stack space
	#
	# Compute base of caller's stack = (sp | mask) + 1
	#         base of new stack      = freeP + size
	#         size of caller's stack = base - sp
        #
	lw	a3, cthread_stack_size
	lw	a1, cthread_stack_mask		
	add	a0, a0, a3
	or	a0, a0, a1
	addi	a0, 1			# a0 = base of new stack
	or	a1, sp, a1
	addi	a1, 1			# a1 = base of caller's stack
	sub	a2, a1, sp		# a2 = size of caller's stack
	sub	a0, a0, a2		# a0 = base of new stack - size

	#
	# Call memcpy to copy caller's stack
	#
	sw	a0, 12(sp)	# save new stack pointer
	jal	memcpy		# char *memcpy (char *dest, char *src, int len)
	move	a1, sp

	#
	# Change sp to point to new stack
	#	
	lw	ra, 16(sp)
	lw    	sp, 12(sp)	
	j	ra
	addi	sp, 20		# 2 saved regs, 3 arguments

END(imove_stack)
