/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */

/*
 * 27 Jun 91 (mjz): this "last" buisness is all messed up.  It's not
 *		really necessary. last === ld->next, and when last==q
 *		it's time to add new elements.  Toss last.
 *  8 Jun 92 (mjz): Today's hard-to-find memory overwrite: initialized
 *		extq with WRONG SIZED ELEMENT.  BARF.  Only mattered if
 *		Q was rearranged due to push, I think... sigh.
 * 11 Aug 92 (was): added clock_t type
 *  9 Mar 94 (was): added incarn to queued elements
 */

#include "midway_types.h"        /* for clock_t */


/* general queue structure */

typedef struct ourqe {		/* generic queue elment */
  struct ourqe *next;		/* just a next pointer in common place */
} ourqe_t;
typedef struct ourq {
  ourqe_t	*ld;		/* last with data */
  ourqe_t	*q;		/* the q */
} ourq_t;

typedef ourq_t threadq_t;	/* thread id queue */
typedef ourq_t extq_t;		/* an external location queue */

typedef struct threadqe {
  struct threadqe *next;	/* next element - always first */
  u_long	tid;		/* thread id */
} threadqe_t;

typedef struct extqe {
  struct extqe *next;		/* next element */
  u_long	host;		/* host that is requesting */
  u_long	tid;		/* thread id on that host */
  u_long	tranid;		/* transaction id too...  */
  clock_t	clock;		/* clock value (OF LOCK) */
  clock_t	incarn;		/* lock's incarnation */
} extqe_t;

/* following MACRO_BEGIN stuff shamelessly copied from ecc's cthreads.h */
#ifndef MACRO_BEGIN
# ifdef lint
    int NEVER;
# else  lint
#   define NEVER 0
# endif lint
# define MACRO_BEGIN	do {
# define MACRO_END	} while (NEVER)
#endif MACRO_BEGIN

/*
 * initialized: ourq structure with one element of proper type hanging off
 *
 * queue has values from q->next to ld, and spare elements from ld to q.
 *   (queue size may grow, but never shrinks)
 *
 * queue is empty with q and ld pointing to the same element; next of that
 *   element points to itself initially...
 * 
 * queue needs expansion if ld->next == q.
 *
 * pop gives element value, then cycles cell to the end.
 *
 */

/* errors from calloc? */
#define mid_threadq_init(x) \
	MACRO_BEGIN \
           (x)->q = (ourqe_t *) calloc(1,sizeof(threadqe_t)); \
           (x)->ld = (x)->q->next = (x)->q; \
	MACRO_END

#define mid_extq_init(x) \
	MACRO_BEGIN \
           (x)->q = (ourqe_t *) calloc(1,sizeof(extqe_t)); \
           (x)->ld = ((x)->q)->next = (x)->q; \
	MACRO_END

#define mid_ourq_empty(x) ((x)->q == (x)->ld)
#define mid_threadq_empty mid_ourq_empty
#define mid_extq_empty mid_ourq_empty

/* stdc stuff? */
extern u_long mid_threadq_top();
