/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * socket_iface.c: provide a veneer over sockets so we can have a common
 *	interface for sends and receives between the sockets interface
 *	and the ATM interface.
 *
 *  9 Oct 93 (mjz): new
 * 11 Nov 93 (mjz): add wantreply (does nothing for now)
 */
#include <sys/types.h>		/* u_long & friends */
#include <netinet/in.h>		/* sockaddr_in */
#include <sys/socket.h>		/* why not? */
#include <cthreads.h>		/* comm.h dependencies */
#include <stdio.h>		/* NULL */
#include "midway_types.h"	/* global midway type definitions */
#include "comm.h"		/* NET_RECVFROM */
#include "pbuf.h"		/* pbuf definitions */
#include "socket_iface.h"	/* external fn definitions */


extern int midway_udpsock;	/* set by CommOpen in comm.c */

int midway_socket_rcv(char **buf, 
		      int *buflen, 
		      u_int *rcvtime,
		      struct sockaddr *sin)
{
  int		sinlen;		/* length of sin */
  int		rc;		/* return code */

  /*
   * translate to socket recvfrom
   *
   * wait until a message is received
   */
  /* allocate buffer for this thread */
  *buf = mid_pbuf_get ();
  *buflen = mid_max_pbuf_size;
  if (*buf == NULL) {
    StatusMsg("socket_rcv: insufficient buffer space, dying.");
    cthread_exit((void *)101);
  }

  sinlen = sizeof(struct sockaddr_in); /* really, not sockaddr */
  rc = NET_RECVFROM(midway_udpsock, *buf, *buflen, 0,
		    sin, &sinlen);
  *rcvtime = 0;			/* XXX */

  return(rc);
} /* midway_socket_rcv */

int midway_socket_sendto(
			 int s,
			 char *msg,
			 int len,
			 int flags,
			 struct sockaddr *to,
			 int tolen,
			 int dest
			 )
{
  /* just ignore "dest" */
  int rc = NET_SENDTO(s, msg, len, flags, to, tolen);
  return(rc);
} /* midway_socket_sendto */

int midway_socket_sendmsg(
			  int s,
			  struct msghdr msg[],
			  int flags,
			  int dest
			  )
{
  int rc = NET_SENDMSG(s, msg, flags);
  return(rc);
} /* midway_socket_sendmsg */

int midway_socket_wantreply(
			    comm_filter filt,
			    int cpti
			    )
{
  return(-1);			/* we don't handle */
} /* midway_socket_wantreply */
  
