/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * Simple timer package
 *
 * 17 Aug 92  (bershad)		Create
 *  3 Oct 92 (mjz): remove chatty "in timer_init" message
 * 20 Oct 92 (was): added midway_task_self for mach_task_self()
 *  4 Mar 94 (mjz): undo midway_task_self; called only once, and
 *		I want to use it before midway_task_self is set up.
 */

#include <sys/types.h>
#include <mach.h>
#include <mach/message.h>

static mach_port_t timeoutport = 0;
static mach_msg_header_t msgh;	    
extern task_t midway_task_self;	/* mach_task_self() */

timer_pause_ms(ms)
{
    if (!timeoutport)  {
	timer_init();
    }
    return mach_msg(&msgh,
		    MACH_RCV_MSG|MACH_RCV_TIMEOUT,	/* option */
		    0,			/* send size */
		    0,			/* receive size */
		    timeoutport,	/* port */
		    ms,
		    MACH_PORT_NULL);
		    
}


kern_return_t
timer_init()
{
    kern_return_t krc;
    krc = mach_port_allocate(mach_task_self(),
			     MACH_PORT_RIGHT_RECEIVE, &timeoutport);
    if (krc != KERN_SUCCESS)  {
        mach_error("mach_port_allocate", krc);
    }
    msgh.msgh_local_port = MACH_PORT_NULL;
    msgh.msgh_remote_port = MACH_PORT_NULL;
    return krc;
}
