/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
#include <mach.h>

/*
 * vm_shared_base is start of dynamic shared area
 * vm_shared_end is end.
 */
extern vm_offset_t vm_shared_base;
extern vm_offset_t vm_shared_end;

/*
 * vm_dirty_base is start of dynamic dirty bit area
 * vm_dirty_end is end.
 */
extern vm_offset_t vm_dirty_base;
extern vm_offset_t vm_dirty_end;
extern clock_t     vm_dirty_value;
extern vm_offset_t vm_dirty_free_spaceP;
extern vm_offset_t vm_dirty_free;


/*
 * vm_static_base is start of user's static data
 * vm_static_end is highest address for user's static data.
 * Note: The area may be discontiguous, since the mips loader
 * allocates static data in a variety of areas. See midway_start.s
 * and lock.c for more details.
 */
extern vm_offset_t vm_static_base;
extern vm_offset_t vm_static_end;


#ifdef PAGE_BASED
/*
 * vm_paged_based is start of paged (ie shared) area
 * Shared & Dirty areas are not used.
 */
extern vm_offset_t vm_paged_base;
extern vm_offset_t vm_paged_end;
extern vm_offset_t vm_paged_free_spaceP;
extern vm_offset_t vm_paged_free;

#endif


/*
 * vm_private_base is start of the application's private area.
 * It contains the application's thread's stack
 * along with the heap.
 */
extern vm_offset_t vm_private_base;
extern vm_offset_t vm_private_end;
extern vm_offset_t vm_private_stackP; 		/* ptr to base of application's stack */
extern vm_offset_t vm_private_free_spaceP;
extern vm_offset_t vm_private_free;

/*
 * vm_locktable_base is the start of midway's locktable.
 */
extern vm_offset_t vm_locktable_base;
extern vm_offset_t vm_locktable_end;
extern vm_offset_t vm_locktable_free;



/*
 * Define macro to convert an address to a logical page number
 */
extern vm_size_t vm_page_shift;
#define ADDR_TO_PAGENO(_addr) ((_addr) >> vm_page_shift)
#define PAGENO_TO_ADDR(_addr) ((_addr) << vm_page_shift)


/*
 * verify_region:
 *	Simple verification function to determine if an accessed page
 *	is within a known region. The routine get_shmalloc_info will
 *	return with KERN_SUCCESS if the offset is within a known region
 *	(possibly after going out and discovering the region information
 *	 from another host).
 */
#define IN_SHARED(x) (((x) >= vm_shared_base) && ((x) <= vm_shared_end))
#define IN_DIRTY(x) (((x) >= vm_dirty_base) && ((x) <= vm_dirty_end))
#define IN_PRIVATE(x) (((x) >= vm_private_base) && ((x) <= vm_private_end))
#define IN_LOCKTABLE(x) (((x) >= vm_locktable_base) && ((x) <= vm_locktable_end))

#define IN_DIRTY_EXTRA_SPACE	(sizeof(clock_t)*3) /* extra space req'd for loops */

/*
 * prototypes
 */
kern_return_t vm_shared_init();
kern_return_t vm_allocate();	/* we supply our own thankyou. */






