/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
#include <mach.h>
#include <stdio.h>
#include <syslog.h>
#include "smexecd_types.h"	/* char255 */

extern boolean_t midway_xpager_debug;
extern boolean_t midway_xpager_use_syslog;

char *
ErrString(status)
    int status;
{
    char *mach_error_string();

    if (status == -1) {
	return "Unix Error?";
    } else {
	return mach_error_string(status);
    }
}

static boolean_t
using_syslog()
{
    static boolean_t openlog_done = FALSE;

    if (! midway_xpager_use_syslog) {
	return FALSE;
    }
    if (midway_xpager_debug) {
	return FALSE;
    }
    if (! openlog_done) {
	openlog("midway xpager", LOG_PID | LOG_CONS, LOG_DAEMON);	
	openlog_done = TRUE;
    }
    return TRUE;
}

fret(fmt, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10)
    char *fmt, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9, *a10;
{
    char255 smsg;
    if (using_syslog()) {
	syslog(LOG_ERR, fmt, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10);
    } else {
	sprintf(smsg, fmt, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10);
	StatusMsg(smsg);
    }
}

dbg(fmt, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10)
    char *fmt, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9, *a10;
{
    char buf[1024];
    strcpy(buf, "dbg:");
    strcat(buf, fmt);    
    fret(buf, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10);
}

statusmsg(fmt, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10)
    char *fmt, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9, *a10;
{
    char buf[1024];
    strcpy(buf, "dbg:");
    strcat(buf, fmt);    
    fret(buf, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10);
}



mumble(fmt, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10)
    char *fmt, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9, *a10;
{
    if (midway_xpager_debug) {
	fret(fmt, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10);
    }
}

#if !defined(FAST_UDP)
panic(fmt, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10)
    char *fmt, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8, *a9, *a10;
{
    char buf[1024];

    strcpy(buf, "panic: ");
    strcat(buf, fmt);
    fret(buf, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10);
    exit(1);
}
#endif

/*
 * Various utility functions
 */
void
botch(p)
    char *p;
{
    printf("botch: %s\n", p); /* gag. Don't use printf */
    printf("botch: SPINNING! (pid %d)\n",getpid());
    while(1) {
	cthread_yield();
    }
    /*    exit(-1); */
}

char * 
botch_message(char *file, int line)
{
  static char botch_msg[256];

  sprintf(botch_msg, "%s:%d", file, line);
  return (botch_msg);
}
