/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1989 Carnegie-Mellon University
 * All rights reserved.  The CMU software License Agreement specifies
 * the terms and conditions for use and redistribution.
 */
/*
 * HISTORY
 * $Log: xpager_init.c,v $
 * Revision 1.2  1992/10/24  21:43:21  wsawdon
 * Fixed 2 bugs in external pager.
 *   Page data passed back on a data_request must be page aligned.
 *   Always give R/W access even if R/O requested.
 *
 * Revision 1.1  1992/10/01  20:00:53  mattz
 * External Pager for the Midway runtime.  Keeps track of shared storage.
 *
 */

#include <mach.h>
#include <mach/message.h>
#include <mach/notify.h>
#include <cthreads.h>
#include <servers/netname.h>
#include "xpager.h"


static char *servername = "xpager";

mach_port_t xpager_port;
cthread_t xpager_thread;

extern boolean_t XPager_server();

static void
unix_bogosity(dummy)
    int *dummy;
{
    while (1)  {
	pause();
    }
}


xpager_server(server_set)
    mach_port_t server_set;
{
    kern_return_t status;

    /* we are cthread_fork'd; for now wire all cthreads with 3.0 threads*/
    cthread_wire();

    for (;;){
	status = mach_msg_server(XPager_server, 8192, server_set);
	if (status != MACH_MSG_SUCCESS)	{
	    fret("mach_msg_server returns ! SUCCESS: %d: %s\n", status,
		 mach_error_string(status));
	}
    }
}

kern_return_t
xpager_init(announce)
    int announce;
{
    kern_return_t status;
    mach_port_t notify_port;
    mach_port_t server_set;

    if (announce)  {
	status = netname_look_up(name_server_port, "", servername,
				 &xpager_port);
	if (status == NETNAME_SUCCESS) {
	    panic("An instance of %s is already running.\n", servername);
	}
	status = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &xpager_port);
	if (status != KERN_SUCCESS) {
	    panic("main: mach_port_allocate: %s\n", ErrString(status));
	}
	status = netname_check_in(name_server_port, servername, 
				  mach_task_self(), xpager_port);
	if (status != NETNAME_SUCCESS) {
	    panic("main: netname_check_in: %s\n", ErrString(status));
	}
    
	status = task_get_special_port(mach_task_self(), 2, &notify_port);
	if (status != KERN_SUCCESS) {
	    panic("task_get_special_port: can't get notify port\n");
	}

	status = mach_port_allocate(mach_task_self(),
				    MACH_PORT_RIGHT_PORT_SET, &server_set);
	if (status != KERN_SUCCESS) {
	    panic("main: port_set_allocate [SET]: %s\n", ErrString(status));
	}
	status = mach_port_move_member(mach_task_self(),
				       notify_port,
				       server_set);
	if (status != KERN_SUCCESS) {
	    panic("main: mach_port_move_member: %s\n", ErrString(status));
	}
	status = mach_port_move_member(mach_task_self(),
				       xpager_port,
				       server_set);
	if (status != KERN_SUCCESS) {
	    panic("main: mach_port_move_member(2): %s\n", ErrString(status));	
	}
	xpager_thread = cthread_fork(xpager_server, server_set);
    }

}


    

