/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1989 Carnegie-Mellon University
 * All rights reserved.  The CMU software License Agreement specifies
 * the terms and conditions for use and redistribution.
 */

#include <mach.h>
#include <mach/message.h>
#include <mach/notify.h>
#include <servers/netname.h>



/*
 *  The object is normally in READY state.
 *  After a clean or a flush, it moves to FLEANING
 *  until all pages have been flushed or cleaned,
 *  at which point it moves to FROZEN.
 *  When in FROZEN state, a file may be resumed,
 *  in which case the object moves back to READY,
 *  or (if the file was flushed instead of cleaned),
 *  the object will be destroyed. The DEAD state
 *  is used only when returning from filemap_close
 *  so that the reply message can be sent before
 *  the object thread exits.
 */
#define	O_STATE_READY		0
#define	O_STATE_FLEANING	1
#define	O_STATE_FROZEN		2
#define	O_STATE_DEAD		3

extern boolean_t	midway_xpager_may_cache;
extern mach_port_t		xpager_port;

typedef struct object	*object_t;
typedef struct request	*request_t;

#define	OBJECT_NULL	((object_t) 0)


typedef struct max_msg {
    mach_msg_header_t header;
    char data[8192 - sizeof(mach_msg_header_t)];
} max_msg_t;

struct request {
    vm_offset_t		offset;
    vm_size_t		length;
    vm_prot_t		desired_access;
    boolean_t		needs_data;
    request_t		next;
};

struct object {
    mutex_t		mu;
    char		state;
    boolean_t		flean_was_flush;
    mach_port_t		memory_object;
    mach_port_t		memory_control;
    char		*buf;
    request_t		r_head;
    request_t		r_tail;
    vm_size_t		max_req;	/* xxx should be max_read, max_write */
    object_t		next;
    xpager_request_t	request;
    xpager_control_t	control;
    backing_store_t     backing_store;    
};

extern object_t	object_lookup();
extern object_t	object_create();

extern void object_lock();
extern void object_unlock();

    



			   


