/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 *  4 Dec 91 (mjz): fixed a couple of bugs of the "off by one" variety in
 *		ParseFile; long comments were causing the program to be
 *		read incorrectly due to the semantics of fgets.
 * 16 Dec 91 (mjz): attempt to combat startup anomoly 'cause of contention
 *		for this processor.  
 *		1. Move startup code before fire-up code
 * 11 Aug 92 (was): added clock_t type
 *  3 Oct 92 (mjz): remove bnb debugging {bnb modif'd for passing ports via NS}
 *  4 Oct 92 (mjz): incarnation #, de-verbosify (;-)), smexecd per uid
 *  9 Oct 92 (was): converted network calls to macros NET_xxxx
 * 27 Oct 92 (was): added macro for smexec_debug
 * 16 Nov 92 (mjz): new death_thread; calls to ss_exit.
 *  8 Jul 93 (mjz): FAST_UDP network_init needs parameter
 */
#include <mach.h>
#include <mach_error.h>
#include <cthreads.h>
#include <servers/netname.h>
#include <ctype.h>		/* for isspace */
#include <strings.h>		/* for rindex */
#include <sys/param.h>		/* MAXHOSTNAMELEN */
#include <sys/types.h>		/* for socket.h, in.h */
#include <sys/socket.h>		/* socket constants */
#include <netinet/in.h>		/* sockaddr_in, and company */
#include <stdio.h>
#include <errno.h>
#include <signal.h>		/* for death */
#include "runtime/memory/midway_types.h"       /* global midway type definitions */
#include "smexecd.h"
#include "sm.h"			/* SMNAME, UDPPORTSTART... */

#include "runtime/memory/comm.h"	/* for HOST_OP operations in main() */
#include "runtime/memory/pbuf.h"	/* for MAX_DATA_SIZE */

#define KERNEL 1		/* HACK: workaround for CMU include file incompatabilty */
#include <sys/time.h>
#undef KERNEL			/* HACK */


extern int errno;		/* why not in header file? */
extern char *sys_errlist[];	/* printable errno messages */
extern int sys_nerr;		/* # of those messages */

/* following here so death_hand can use... */
static COMMMSGP	cmp;		/* for HOST_OP reply... */
static int num;			/* num = #things in hostinfo */
static int rsock;		/* for host_op thingys */

static u_long smexec_incarnation; /* incarnation # for this execution */

/* disabling these checks does not save us much, and gives me a headache.
 * --mjz 10 apr 93 */
/*#ifdef BENCHMARK
 *# define SMEXEC_DEBUG 0
 *#else
 */
# define SMEXEC_DEBUG smexec_debug
/*
 *#endif
*/
int smexec_debug = 0;		/* hmmm */

typedef struct argsent {
	char *host;		/* name of host */
	char *args;		/* argv for program on host */
	mach_port_t smexecp;		/* port for smexec server on host */
    } ARGSENT;



mach_port_t	status_server_port;	/* give to kids */
mach_port_t status_server_set;	/* port set; status_server() needs */
mach_port_t midwayux_server_port;   /* give to kids */
mach_port_t midwayux_server_set;    /* for the server */
cthread_t	ss_thread;	/* mostly for debugging */


/*
 * fire_up(): start the program running on all the named hosts with the
 *		given arguments.
 */
fire_up(prog,hi,num,ssp,hlist,rhost,rport,uxp)
     char	*prog;		/* name of program to fire up */
     ARGSENT	*hi;		/* the hostinfo structure */
     int	num;		/* number of entries in hi */
     mach_port_t	ssp;		/* status server port */
     char	*hlist;		/* hname info from hi, but compressed */
     u_long	rhost;		/* ip address */
     u_short	rport;		/* UDP port */
     mach_port_t	uxp;		/* I/O (UniX) server port */
{
    char255	program;
    char255	progargs;
    kern_return_t retcode;
    int		i;
    u_long	hlistCnt;

    hlistCnt = strlen(hlist);
    strcpy(program,prog);	/* there's now two copies one by us on MiG */
    for (i=0; i<num; i++) {
/*	int xx; bnb debugging, mjz is removing seems real stupid here */
	strcpy(progargs,hi[i].args);
/*	smexec_ping(hi[i].smexecp,  &xx); mjz removing */
	if (SMEXEC_DEBUG)
	  printf("Calling start_proc %s:%s(%s)\n", hi[i].host, program, progargs);
        retcode = start_proc(hi[i].smexecp, ssp, uxp, rhost, rport,
			     hlist, hlistCnt, num, i+1, 
			     smexec_incarnation, program, progargs);
        if (retcode == KERN_SUCCESS) {
	  if (SMEXEC_DEBUG)
	    printf("start_proc() was successful!\n");
	}
        else 
	  mach_error("Error from start_proc is",retcode);
      } /* for i = 0 to num-1 */
} /* fire_up */

/*
 * ParseFile(): pick apart program, and host/argstring pairs.
 */
static int ParseFile(who,fn,num,maxl,progname,hi)
     char	*who;		/* for error messages */
     char	*fn;		/* filename */
     int	*num;		/* number of hosts (out) */
     int	*maxl;		/* max input length (for hahas) (out) */
     char	**progname;	/* program to execute (out) */
     ARGSENT	**hi;		/* the hostinfo array (out) */
{
     int	maxlen;		/* so we don't need to always dereference */
     ARGSENT	*hostinfo;	/* ditto */
     int	nl,len,com;	/* for parsing */
     int	i,j;		/* joe random integers */
     char	*argheap;	/* where data for hostinfo comes */
     FILE	*fp;		/* will be open file */
     register int c;		/* char but on getc */
     register char *cp;		/* will be pointer into argheap */

     if ((fp = fopen(fn,"r")) == NULL) {
	fprintf(stderr, "can't open file %s\n", fn);
	perror("fopen");
	fprintf(stderr,"...for usage, type `%s -h'.\n",who);
	return(2);
     }
     /* want max length of input, and # of host entries.  Unfortunately,
	with Unix i/o one must dogpaddle through file to get this...

	Alternative would be (1) fix size (2) use linked list.
	(2) is bad 'cause will be using these #'s as indexes into
	structures later...
      */
    nl = 1;		/* pretend we start at a nl */
    maxlen = 0;		/* none yet */
    len = 0;
    com = 0;		/* is comment */
    *num = -1;		/* no valid lines yet either */
    /* NOTE: Before today, I was not counting the length of lines in
     * comments.  This means I may allocate less space, if the only long
     * lines are comment lines.  Unfortunately, it also means that the
     * fgets to get the progname may end up not getting an entire line,
     * causing junk to be assigned to the program name (since the rest
     * of the long comment line does not start with a comment character).
     * Since fgets does not return the length of the input it returns,
     * I would have to scan the input to see if there was a newline in it,
     * and repeatedly fgets until I got the newline.  Bletch.  This would
     * make the code less elegant, and I also don't feel like scanning
     * each input line (I guess I would only have to scan the comment
     * lines).  In any event, I'll take the increased memory use for
     * ease of coding later.  We now count the length of comment lines
     * in maxlen.  4 dec 91, mjz.
     */
    for (c = getc(fp); c!=EOF; c = getc(fp)) {
	if (nl){
	    if (c == '\n') continue;
	    if ((c == '*') || (c == '#') || (c == '!')) {
	        com = 1;	/* A comment! */
	    }
	    len++;		/* not newline, start length (must be 1 now) */
	    nl = 0;		/* last not a newline */
	} else {		/* not on nl */
	    if (c == '\n') {
		nl = 1;
		if (len > maxlen) maxlen = len;
		if ((len > 0) && !com) (*num)++;	/* seen a new line */
		len = 0;		/* reset */
		com = 0;
	    } /* newline */
	    else {
		len++;		/* inc len if not nl (normal case) */
	    }
	} /* not on nl */
    } /* for */
    /* take care of stopping w/o seeing eol */
    if (len > maxlen) maxlen = len;
    if (len > 0) (*num)++;
    /* error if *num < 1 (no hosts) */
    if (*num == -1) {
	fprintf(stderr,"Error: %s contains no information!\n", fn);
	printf("...for usage, type `%s -h'.\n",who);
	return(3);
    }
    if (*num == 0) {
	fprintf(stderr,"Error: %s contains no host/argument lines!\n", fn);
	printf("...for usage, type `%s -h'.\n",who);
	return(3);
    }

    if (SMEXEC_DEBUG)
      printf("%s: %d data lines, max size is %d.\n", fn, *num, maxlen);/*mjz*/

    /* alloc hostinfo array */
    hostinfo = (ARGSENT *) malloc(*num*sizeof(ARGSENT));
    if (hostinfo == (ARGSENT *) 0) {
	perror("malloc of hostinfo array failed");
	return(4);
    }
    /* alloc "heap" for host / argv data */
    argheap = (char *) malloc((*num)*(maxlen+1));
    if (argheap == NULL) {
	perror("malloc of argheap failed");
	return(4);
    }
    *progname = (char *) malloc(maxlen+2);
    /* +2 = +1 for \n and +1 for \0 inserted by fgets (else chops off
     * longest line right before \n, and get \n for name.  Bletch. */
    if (*progname == NULL) {
	perror("malloc of progname failed");
	return(4);
    }

    /* time to parse the donuts */
    rewind(fp);			/* start at square one */
    /* get progname; loop guaranteed to terminate since *num > 0 */
    do {
	fgets(*progname,maxlen+2,fp);
    } while((**progname == '*')||(**progname == '!')||
	    (**progname == '#'));
    cp = rindex((*progname),'\n');
    if (cp) 
      *cp = '\0';		/* remove trailing \n */

    /* ok.  Now, for each line:
     *   1. skip if comment
     *   2. eat whsp
     *   3. eat host (until whsp)
     *   4. skip whsp
     *   5. eat args
     */
    i = 0;			/* index into hostinfo array */
    cp = argheap;		/* where the good stuff goes */
    c = '\0';			/* prime pump */

    while (c != EOF) {
      while(isspace(c=getc(fp))); /* skip over whitespace */
      if (c == EOF) break;	/* done! */
      if ((c == '#') || (c == '*') || (c == '!')) { /* comment! */
	while((c != EOF) && (c != '\n'))
	  c = getc(fp);		/* eat comment */
	if (c == EOF) break;
	continue;
      } /* comment */

      hostinfo[i].host = cp;	/* found start of host */
      for (; (c != EOF) && (!isspace(c)); c = getc(fp))
	*cp++ = (char) c;	/* copy host name  */
      *cp++ = '\0';		/* null terminate */

      if ((c == EOF) || (c == '\n')) { /* no arguments to follow */
	hostinfo[i].args = cp;
	*cp++ = '\0';
	i++;			/* finished with this host */
	if (c == EOF) break;
	continue;
      }

      /* ok -- skip whitespace before args */
      /* can't use isspace() -- want to catch newlines */
      while ((c==' ')||(c=='\t'))
	c = getc(fp);

      if ((c == EOF) || (c == '\n')) { /* no arguments to follow */
	hostinfo[i].args = cp;
	*cp++ = '\0';
	i++;			/* finished with this host */
	if (c == EOF) break;
	continue;
      }

      hostinfo[i].args = cp;	/* found start of args */
      /* copy through end of line */
      for (; (c != EOF) && (c != '\n'); c = getc(fp))
	*cp++ = (char) c;	/* copy args  */
      *cp++ = '\0';		/* null terminate  */
      i++;			/* prepare for next host */
    } /* end while !eof */

    if (i != *num) {
      printf("Number of entries in hostinfo (%d) does not match expected number (%d)\n", i,*num);
      printf("...using real number of entries in hostinfo\n");
      *num = i;
    }
    if (SMEXEC_DEBUG) {
      printf("progname=`%s'\n", *progname);
      for (j=0; j<i; j++) {
	printf("%d: h=`%s' a=`%s'\n", j, hostinfo[j].host, hostinfo[j].args);
      }
    }

    /* set return values */
    *hi = hostinfo;
    *maxl = maxlen;
    return(0);
} /* ParseFile */

/*
 * death_hand: death handler.  If we receive a ^C, or a killall message,
 * 	then we attempt to kill all our processes...
 */
static int death_hand_invoked = 0;
static death_hand()
{
    struct sockaddr_in sin;
    int rc;
    kern_return_t krc;
    HREPLYMSGP hrp;
    int i, numloop;

    /* tell our internal waiter that the signal handler is doing work */
    death_hand_invoked = 1;

    ss_exit(); /* release anyone waiting, should kill others */
    /* we should be circumventing mig with ss_exit */

    /* send out messages anyway */

    if ((cmp == (COMMMSGP) 0) || (num == 0))
    exit(199); /* nothing more we can do */

    /* prep sending socket */
    sin.sin_family = AF_INET;
    bzero(sin.sin_zero, sizeof(sin.sin_zero));

    /* hostop stuff follows immediately after COMMMSG.  Coerce cmp to
       a character pointer so addition is in bytes. */
    hrp = (HREPLYMSGP) ((char *) cmp + sizeof(COMMMSG));

    /* and fields in message  (just in case) */
    cmp->version = M_PROTOVER;
    cmp->incarnation = smexec_incarnation;
#ifdef CALLOC_NOTUSED
    cmp->source = 0;
    cmp->id = 0;
    cmp->clock = NULL_CLOCK_VALUE;
    cmp->lock = 0;
#endif CALLOC_NOTUSED
    cmp->rid = 0;
    cmp->tid = 0xdeadbeef; /* why not? */
    /* following should only need be > max # of host op messages sent out */
    cmp->clock = MAX_CLOCK_VALUE; /* max clock for now */
    cmp->flags = M_LASTPACK;
    cmp->type = M_SEPPUKU;
    ZERO_ADDR_VECTOR(cmp->nacks);

    /* Strategy: send suicide messages to all hosts, wait, and do it again
     * for two more times.  Then wait a little bit more (for status messages
     * to arrive, etc) and kill ourselves.
     */

    for (numloop = 0; numloop < 3; numloop++) {
	fprintf(stderr,"Sending suicide message to");
	for (i=0; i<num; i++) {	/* for all hosts */
	    if (hrp[i].udpport) {
		fprintf(stderr, " %lu", i);
		sin.sin_addr.s_addr = hrp[i].ipaddr;
		sin.sin_port = hrp[i].udpport;
		rc = NET_SENDTO(rsock, cmp, sizeof(COMMMSG), 0, 
				&sin, sizeof(sin));
		if (rc == -1) {
		    fprintf(stderr, "\nfailure sendto suicide msg to %d: %s (%d)\n",
			    i+1, ((errno>sys_nerr)?"unknown":sys_errlist[errno]),
			    errno);
		    /* continuing anyway... */
		}
	    } else { /* no info */
		fprintf(stderr, " *%lu*", i);
	    }
	} /* for all hosts */
	putc('\n',stderr); fflush(stderr);
	timer_pause_ms(1000);
    } /* for loop 3 times */

    /* any reason we should timeout here? */
    /* timeout here is so ending status server messages can be redirected
       back... */
    timer_pause_ms(2000);

    exit(0); /* bye bye */
}   /* death_hand */

/* allow us to die on midway_exit() */
static int
death_thread()
{
  cthread_wire();
  ss_waitfor_exit();		/* will circumvent MiG; no port needed */

  if (death_hand_invoked == 0) {
    fprintf(stderr, "==> waiting two seconds for all threads to die...\n");
    timer_pause_ms(2000);	/* allow final messages to appear */

    exit(0);
  }
  /* else called internally, someone else will call exit */
}

/* mach 3 bogosity handler */
int
mach3bogus(arg)
     any_t arg;
{
  cthread_wire();		/* for Mach 3 cthreads -- keep kernel thread */
  while(1) {			/* my favorite kind of loop! */
    sigpause(0);		/* wait around in UX Server */
#ifdef VERBOSITY
    fprintf(stderr,"smexec: sigpause terminated.\n"); /* debug */
#endif VERBOSITY
  }
} /* mach3bogus */



initialize_server_ports()
{
    extern int ss_server();
    extern int midwayux_server();

    /* create status server port and port set */
    /* check in status server port for debugging */   
    if (allocate_server_ports(&status_server_port,&status_server_set,
			      "SMstatusServer")) {
	exit(4); /* bail out */
    }
    ss_thread = spawn_server_thread(ss_server, status_server_set, 4096);

    /*
     * I/O server.
     */
    if (allocate_server_ports(&midwayux_server_port, &midwayux_server_set, 0)) {
	exit(5); /* bail out */
    }
    cthread_detach(spawn_server_thread(midwayux_server, midwayux_server_set, 4096)); 
}




/*
 * smexec <-h|?|[control-file-name]>
 *
 * -h, -? ?: usage
 * control-file-name:
 *	Reads in file name to exec  (EXEC as whom??? --mjz)
 *	Read in host / argstring for host pairs
 *      fire up programs on all hosts
 *
 * control file is of format:
 *	* this is a comment
 *	! so is this
 *      # so is this
 *	/afs/cs.cmu.edu/file-to-exec
 *	host1<whitespace>args to pass to host 1
 *      host2<whitespace>args to pass to host 2
 *      ...
 *      hostn<whitespace>args to pass to host n
 *
 * We artificially limit us to 100 hosts, #chars/???
 */

main(argc,argv)
    int argc;
    char **argv;
{
    ARGSENT	*hostinfo;		/* host info array (eventually) */
    char	*progname;		/* for program name */
    int		maxlen;
    int		rc;		/* return code */
    kern_return_t krc;		/* one for kernel calls */
    int		i;		/* random integer */
    char	*default_fn = "smexec.list";
    char	*fn = default_fn;
    char	*hlist;		/* compact list of hosts */
    char	*cp,*cq;	/* random character pointer */
    struct hostent *hp;		/* for name service */
    extern int	h_errno;	/* for name service */
    char	myname[MAXHOSTNAMELEN+1]; /* for my ip address */
    u_long	myip;		/* my IP address */
    u_short	rport;		/* UDP port for initial rendesvous */
    struct sockaddr_in sin;	/* for addressing said socket */
    struct sockaddr_in from_sin;/* udp sender */

    HREPLYMSGP  hrp;		/* for host entries w/in message */
    u_long	cml;		/* length of entire message... */
    u_long	bcnt;		/* # of hosts that reached this "barrier" */
    struct {			/* to receive HOST_OP requests */
      COMMMSG	c;
      HREPLYMSG h;
    } reqmsg;
    typedef struct {
      u_long	thread;		/* for each host, also must keep */
      u_long	tid;		/* thread & transaction id of req */
    } TTS;
    TTS		*ttp;		/* where to keep them */
    struct sockaddr_in ssin;	/* for sending out host_op replies */
    u_long	u;		/* random index */
    u_long	id;		/* index into ttp,hrp */
    clock_t	myclk;
    int		sinsize;	/* for recvfrom */
    struct timeval tval;	/* for incarnation */
    int		c;		/* for getopt */
    extern int	optind;		/* for getopt */
    extern char	*optarg;	/* for getopt */
    uid_t	uid;
    char	smexecName[256]; /* for building uid specific smexec name */
    int		noUid = 0;	/* don't use uid for smexecd search */
    char	*usage = "\n\n\
Usage: %s <-h|?|[control-file-name]>\n\
\n\
-h, -? ?: usage\n\n\
-d: Turn on smexec initialization+HOST_OP debugging\n\n\
control-file-name:\n\
	Reads in file name to exec  (EXEC as whom??? --mjz)\n\
	Read in host / argstring for host pairs\n\
	fire up programs on all hosts\n\
\n\
control file is of format:\n\
\n\
* this is a comment\n\
! so is this\n\
# so is this\n\
/afs/cs.cmu.edu/file-to-exec\n\
host1<whitespace>args to pass to host 1\n\
host2<whitespace>args to pass to host 2\n\
...\n\
hostn<whitespace>args to pass to host n\n\
\n\
"; /* usage */

    /*
     * 1. verify args
     * 2. Read in & verify progname, hosts & args
     * 3. Get smexec server for each host (fail otherwise)
     * 4. fork new thread to be status server
     * 5. fire up programs on various servers (oops -- barrier/lock port?)
     * 6. wait for death??? restart? kill all on one dying?
     * 7. ???
     */

    cmp = (COMMMSGP) 0;		/* so death handler init correctly */
    num = 0;

    if (SMEXEC_DEBUG)
      printf("argc is %d, argv[1] is `%s'\n",argc, argv[1]); /*mjz*/

#ifdef notdef
   if (argc>2)	{
	printf(usage,argv[0]);
	exit(1);
    }

    if (argc == 2)
	if ((*argv[1] == '?')||(*argv[1] == '-')) {
	    printf(usage,argv[0]);
	    exit(1);
	} else {
	    fn = argv[1];	/* file name */
	}
#else /* notdef -- try new arg parsing*/
    while ((c = getopt(argc, argv, "dhqU")) != EOF) {
      switch(c) {
      case 'd':
	smexec_debug = 1;	/* turn on chattyness - one level for now */
	break;
      case 'q':
	smexec_debug = 0;	/* turn off; JustInCase default is on */
	break;
      case 'U':
	noUid = 1;		/* backwards compatibility for hahas */
	break;
      case 'h':
      case '?':
      default:
	printf(usage, argv[0]);
	exit(1);
	/*NOTREACHED*/
	break;
      } /* switch */
    } /* while */
    if (argc > (optind+1)) {	/* extra goop here */
      printf(usage, argv[0]);
      exit(1);
      /*NOTREACHED*/
    } else if (argc == (optind+1)) { /* single argument */
      fn = argv[optind];
    } /* else were no more arguments */
#endif


#ifdef FAST_UDP
  /*
   * Initialize network library routines
   */
  {
    extern void network_init(char *);
    extern int use_device_syscall;

    if (SMEXEC_DEBUG)
      fprintf(stderr, "smexec_main: initializing fast udp library\n");

    /* set global flag to enable device syscall in library */
    use_device_syscall = 1;

    /* initialize library */
    network_init (".sockets.config");
  }
#endif
  
  
/* Read in & verify progname, hosts & args */
    rc = ParseFile(argv[0], fn, &num, &maxlen, &progname, &hostinfo);
    if (rc != 0) {
      exit(rc);
    }
    
    
    /* Get smexec server for each host (fail otherwise) */
    /* in addition, create list of host names that will be passed to all */
    /* worry about FQDNs?  Not for now... */
    hlist = (char *)malloc((maxlen>255?256:maxlen)*num);
    if (noUid) {
      strcpy(smexecName,SMNAME);
    } else {
      extern uid_t getuid();
      uid = getuid();
      sprintf(smexecName,"%s.%d", SMNAME, (int) uid);
    }
    if (SMEXEC_DEBUG)
      fprintf(stderr,"Using '%s' for name to look up smexecd.\n", smexecName);
    for (i=0,cp=hlist; i<num; i++) {
      krc = netname_look_up(name_server_port, hostinfo[i].host, smexecName,
			    &(hostinfo[i].smexecp));
      if (krc != KERN_SUCCESS) {
	printf("Error in looking up port for SMexec server (host %s) (nm %s)\n",
	       hostinfo[i].host,smexecName);
	mach_error("\tnetname_look_up() returns: ", krc);
	printf("SMexec server may not be running on that host.\n");
	exit(krc);
      }
      for (cq=hostinfo[i].host; *cq; cq++,cp++)
	*cp = *cq;		/* copy host name */
      *cp++ = ' ';		/* separate with blanks */
    } /* for i = 0 to num-1 */
    *cp = '\0';			/* null terminate list */
    if (SMEXEC_DEBUG)
      printf(">>> hlist = `%s'\n",hlist);

    /* find out our IP address, so we can pass it on */
    gethostname(myname,sizeof(myname));
    hp = gethostbyname(myname);
    if (!hp) {			/* error will robinson */
      fprintf(stderr,"Error from gethostbyname(%s): ",myname);
      switch(h_errno) {
      case HOST_NOT_FOUND: fprintf(stderr,"HOST_NOT_FOUND\n"); break;
      case TRY_AGAIN:	   fprintf(stderr,"TRY_AGAIN\n"); break;
      case NO_RECOVERY:	   fprintf(stderr,"NO_RECOVERY\n"); break;
      case NO_ADDRESS:	   fprintf(stderr,"NO_ADDRESS\n"); break;
      default: fprintf(stderr,"unknown! h_errno=%d\n",h_errno); break;
      } /* switch */
      exit(3);
    }
    if (hp->h_addrtype != AF_INET) { /* sanity failure; commit suicide */
      fprintf(stderr,"gethostbyname(%s) returned unknown address family %d\n",
	      myname,hp->h_addrtype);
      exit(3);
    }
    bcopy((char *)hp->h_addr, (char *)&myip, sizeof(myip)); /* get addr */

    /* create udp socket for initial rendesvous */
    rsock = NET_SOCKET(AF_INET,SOCK_DGRAM,0);
    if (rsock == -1) {
      perror("udp socket for rendesvous");
      exit(3);
    }
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY; /* don't bind to particular intfce */
    bzero(sin.sin_zero,sizeof(sin.sin_zero)); /* i've seen picky kernels */
    for (rport = UDPPORTSTART; rport <= UDPPORTEND; rport++) {
      sin.sin_port = rport;
      rc = NET_BIND(rsock,&sin,sizeof(sin));
      if (rc == 0) break;	/* we're done */
      if (errno != EADDRINUSE) { /* failure */
	fprintf(stderr,"-?> failure binding UDP socket to port %d: ",rport);
	perror("");
	exit(3);
      }
    } /* for rport...*/


    initialize_server_ports();


    /*
     * manufacture incarnation number from current time
     */
    gettimeofday(&tval, (struct timezone *) 0 /* no zone */);
    smexec_incarnation = (u_long) tval.tv_sec;

    /*
     * idea: wait for host_op messages, just like a barrier computation.
     * When the last one arrives, send a combined reply to all parties.
     * be prepared to handle duplicates/retransmissions; retransmissions
     * after we've sent out the combined message means the mesage
     * didn't make it, and we should resend the bugger.
     *
     * move fire-up code after this so we don't get host_op retransmissions.
     * 
     */
    /* first, check if we will barf (how barf at this point???) */
    cml = sizeof(COMMMSG) + (num * sizeof(HREPLYMSG));
    if (cml > MAX_DATA_SIZE) {
      fprintf(stderr, 
"size of HOST_OP reply is %d, greater than the amount in a maximum sized\n\
message (%d)... someone needs to write the HOST_OP message splitting code!\n",
	      cml, MAX_DATA_SIZE);
      fflush(stderr);
      goto bailout;		/* uck XXXX (how barf?)*/
    }

    /* OK -- allocate message to combine individual messages */
    cmp = (COMMMSGP) calloc(1, cml);
    if (cmp == (COMMMSGP) 0) {
      fprintf(stderr,
	      "Alloc of COMMMSG for reply failed.  Cannot continue...\n");
      fflush(stderr);
      goto bailout;
    }
    ttp = (TTS *) calloc(1, num*sizeof(TTS)); /* for record keeping */
    if (ttp == (TTS *) 0) {
      fprintf(stderr,
	      "Alloc of transaction record keeping array (ttp) failed.\n");
      fflush(stderr);
      goto bailout;
    }
    /* hostop stuff follows immediately after COMMMSG.  Coerce cmp to
       a character pointer so addition is in bytes. */
    hrp = (HREPLYMSGP) ((char *) cmp + sizeof(COMMMSG));

    /* prepare sending socket */
    ssin.sin_family = AF_INET;
    bzero(sin.sin_zero, sizeof(sin.sin_zero));
    /* and invariant fields in the message */
    cmp->version = M_PROTOVER;
    cmp->incarnation = smexec_incarnation;
#ifdef CALLOC_NOTUSED		/* don't need to do this, used calloc */
    cmp->source = 0;		/* we are host 0 */
    cmp->id = 0;
    cmp->clock = NULL_CLOCK_VALUE;  /* no clock for us, either */
    cmp->lock = 0;		/* no specific lock */
#endif CALLOC_NOTUSED
    cmp->flags = M_LASTPACK;	/* assume one packet for now */
    cmp->type = M_HOST_OP | M_REPLY_BIT; /* a host op reply!!! */
    ZERO_ADDR_VECTOR(cmp->nacks);		/* no acksRequired entries here */

    bcnt = 0;			/* no messages yet */
    myclk = MIN_CLOCK_VALUE;	/* clock for me! */
    sinsize = sizeof(from_sin);
		
    /* ALARM to get to bailout? otherwise, how gracefully quit? */
    signal(SIGINT,death_hand);	/* how about this? :-) */
    /* need the following to catch signals if there is no other
       call into the Unix Server for Mach 3 */
    cthread_detach(cthread_fork(mach3bogus, 0));
    cthread_detach(cthread_fork(death_thread, 0));


    /* start up the programs */
    fire_up(progname,hostinfo,num,status_server_port, hlist, myip, rport,
	    midwayux_server_port);

    /* try a cthread_yeild; although I really want this *task* to relinquish
       the processor... *//* still don't understand why some task should run
			     twice before the other one runs once... */
    cthread_yield();

    /* wait for the host_op messages */
    while(1) {
      /* receive message */
      rc = NET_RECVFROM(rsock, &reqmsg, sizeof(reqmsg), 0, 
		&from_sin,
		&sinsize);
      if (rc == -1) {
	perror("recvfrom HOST_OP message (stopping)");
	goto bailout;
      }

      if (reqmsg.c.version != M_PROTOVER) {
	fprintf(stderr, "invalid protocol version received: %lu", 
		reqmsg.c.version);
	continue;			/* no sense checking for duplicates... */
      }
      /* incarnation check. */
      if (reqmsg.c.incarnation != smexec_incarnation) {
	union {
	  u_long in_addr;		/* gross hack to print inet addr */
	  u_char ca[4];		/* split into bytes */
	} u;
	u_short port = from_sin.sin_port;
	u_short nport = ntohs(port);
	
	u.in_addr = from_sin.sin_addr.s_addr;
	fprintf(stderr, "WARNING: received old %d byte packet from %d.%d.%d.%d port %d(or %d) [incarn 0x%lx != our 0x%lx]",
		rc, u.ca[0],u.ca[1],u.ca[2],u.ca[3],
		port, nport,
		reqmsg.c.incarnation, smexec_incarnation);
	continue;
      } /* incarnation mismatch */

      if ((rc != sizeof(reqmsg))||(reqmsg.c.type != M_HOST_OP)) {
	if ((rc == sizeof(COMMMSG)) && (reqmsg.c.type == M_SEPPUKU)) {
	  /* death stuff */
	  fprintf(stderr, "==> Received KillAll Request from %lu(%lu)\n",
		  reqmsg.c.source, reqmsg.c.id);
	  continue;		/* should never reach here */
	} else {		/* really bogus message */
	  fprintf(stderr, "Bogus HOST_OP message received (size %d not %d).\n",
		  rc, sizeof(reqmsg));
	  if (rc >= sizeof(COMMMSG)) { /* try to decode fields */
	    fprintf(stderr,"- source=%lu id=%lu rhost=%lu rid=%lu lock=%lx\n",
		    reqmsg.c.source, reqmsg.c.id, reqmsg.c.rhost,
		    reqmsg.c.rid,reqmsg.c.lock);
	    fprintf(stderr,"- clock=%lu tid=%lu pacnum=%lu flags=%lx type=%lx\n",
		    reqmsg.c.clock, reqmsg.c.tid, reqmsg.c.pacnum,
		    reqmsg.c.flags, reqmsg.c.type);
	  }
	  fprintf(stderr, "...dropping!\n");
	  fflush(stderr);
	  continue;
	} /* really bogus message */
      } /* not a host op message */
      if (SMEXEC_DEBUG) {
	fprintf(stderr, "received HOST_OP from %d (ip=%lx, prt=%x, tid=%x thd=%x)\n",
		reqmsg.c.source, reqmsg.h.ipaddr,reqmsg.h.udpport,
		reqmsg.c.tid, reqmsg.c.id);
	fprintf(stderr, "bcnt=%d, num=%d\n", bcnt, num);
      }
      id = reqmsg.c.source-1;	/* index at 0 */

      /* if all received, assume re-tx, and retransmit answer to this host */
      if (bcnt == num) {	/* all sites have hit barrier */
	/* rexmit answer */
	if (SMEXEC_DEBUG)
	  fprintf(stderr,"retransmit HOST_OP to %d (tid=%x thd=%x)\n",
		  reqmsg.c.source, ttp[id].tid, ttp[id].thread);
	cmp->rid = ttp[id].thread;
	cmp->tid = ttp[id].tid;
	cmp->clock = myclk;
	myclk = ADJUST_CLOCK (myclk,1);
	ssin.sin_addr.s_addr = hrp[id].ipaddr;
	ssin.sin_port = hrp[id].udpport;
	rc = NET_SENDTO(rsock, cmp, cml, 0, &ssin, sizeof(ssin));
	if (rc == -1) {
	  fprintf(stderr, "failure sendto host_op reply to dest %d: %s (%d)\n",
		  reqmsg.c.source,
		  ((errno>sys_nerr)?"unknown":sys_errlist[errno]), errno);
	  /* continue anyway */
	} 
	else if (rc != cml) {
	  fprintf(stderr, "sendto sent only %d bytes out of %d to %d;\
continuing anyway\n",
		  rc, cml, reqmsg.c.source);
	}
      } /* bcnt == num */
      /* if not all received */
      else {
      /*   incorporate & bump counter if not there before */
	if (hrp[id].index == 0) { /* new entry */
	  hrp[id].index = reqmsg.c.source;
	  hrp[id].ipaddr = reqmsg.h.ipaddr;
	  hrp[id].udpport = reqmsg.h.udpport;
	  ttp[id].thread = reqmsg.c.id;
	  ttp[id].tid = reqmsg.c.tid;
	  bcnt++;
      /*   if all received now, send replies to all hosts (purge input?) */
	  if (bcnt == num) {
	    for (u=0; u<num; u++) { /* for all hosts */
	      if (SMEXEC_DEBUG)
		fprintf(stderr, "About to send HOST_OP reply to %d (tid=%x thd=%x)\n", 
			u+1, ttp[u].tid, ttp[u].thread);
	      cmp->rid = ttp[u].thread;
	      cmp->tid = ttp[u].tid;
  	      cmp->clock = myclk;
	      myclk = ADJUST_CLOCK(myclk,1);
	      ssin.sin_addr.s_addr = hrp[u].ipaddr;
	      ssin.sin_port = hrp[u].udpport;
	      rc = NET_SENDTO(rsock, cmp, cml, 0, &ssin, sizeof(ssin));
	      if (rc == -1) {
		fprintf(stderr, 
			"failure sendto host_op reply to dest %d: %s (%d)\n",
			u+1,
			((errno>sys_nerr)?"unknown":sys_errlist[errno]), 
			errno);
		/* continue anyway */
	      } 
	      else if (rc != cml){
		fprintf(stderr, "sendto sent only %d bytes out of %d to %d;\
continuing anyway\n",
			rc, cml, u+1);
	      }
	    } /* for all hosts */
	  } /* bcnt==num now */
	} /* new entry */
	/* ELSE could send keepalive... */
      } /* if not all received */
    } /* infinite while */
    /* skipped some steps here */

bailout:
    cthread_join(ss_thread);/*mjz*/ /* wait here forever */
} /* main */
