/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * 14 Oct 92 (mjz): wire down threads
 */
#include <stdio.h>
#include <mach.h>
#include <mig_errors.h>
#include <mach/message.h>
#include <strings.h>
#include <cthreads.h>


#define TASK_NOTIFY_PORT 2

allocate_server_ports(sp, sps, name)
    mach_port_t *sp;
    mach_port_t *sps;
    char *name;
{
    kern_return_t krc;
    mach_port_t notify_port;

    /* allocate service port for receiving request messages */
    krc = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, sp);
    if (krc != KERN_SUCCESS) {
	mach_error("mach_port_allocate for io server", krc);
	return 1;
    }

    /* allocate a port set to hold the service port and notify port */
    krc = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_PORT_SET, sps);
    if (krc != KERN_SUCCESS) {
	mach_error("mach_port_set_allocate for io server", krc);
	return 2;
    }

    /* put the service port and notify port into the port set */   
    krc = mach_port_move_member(mach_task_self(), *sp, *sps);
    if (krc != KERN_SUCCESS)  {
	mach_error("Could not initialize port sets", krc);
	return 3;
    }
    krc = task_get_special_port(mach_task_self(), TASK_NOTIFY_PORT, &notify_port);
    if (krc != KERN_SUCCESS)  {
	mach_error("Could not get notify port");
	return 4;
    }
    krc = mach_port_move_member(mach_task_self(), notify_port, *sps);
    if (krc != KERN_SUCCESS)  {
	mach_error("Could not initialize port sets", krc);
	return 3;
    }   

    if (name)  {
	krc = netname_check_in(name_server_port, name, MACH_PORT_NULL, *sp);
	if (krc != KERN_SUCCESS) {
	    mach_error("-?>netname_check_in server port", krc);
	    fprintf(stderr,"-?> ...continuing\n");
	}
    }
    return 0;
}


static struct server_args {
    cthread_fn_t demux;
    mach_port_t port;
    int msg_size;
};
    
	

static
server_thread_root(arg)
    any_t arg;
{
    kern_return_t krc;
    struct server_args *sa = (struct server_args*)arg;
    cthread_wire();		/* nail this down; helps debugging */
    krc = mach_msg_server(sa->demux, sa->msg_size, sa->port);
    if (krc != KERN_SUCCESS)  {
	mach_error("server_thread_root", krc);
    }
    free(sa);
}

    
cthread_t
spawn_server_thread(server_proc, server_port, server_msg_size)
    cthread_fn_t server_proc;
    mach_port_t server_port;
    int server_msg_size;
{
    struct server_args *sa;
    
    sa = (struct server_args*)malloc(sizeof(struct server_args));
    if (!sa)
       return 0;
    sa->demux = server_proc;
    sa->port = server_port;
    sa->msg_size = server_msg_size;
    return cthread_fork(server_thread_root, sa);
}


    
    
    
