/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * 15 Dec 91 (mjz): added "I'm up" warm fuzzy message
 * 29 Jan 92 (mjz): Current mach3 bogosity: a process will not receive a
 *		signal unless it makes a kernel call (which will talk
 *		to the Unix server).  To get around this, I will start
 *		a thread that simply does a sigpause(0), which will cause
 *		that thread to hang in the Unix server until something
 *		happens.  Whoopee!  This means smexec must now be
 *		linked with the cthreads library...
 *	NOTE: I think this will work, 'cause a test of sleep() with two
 *		threads "does the right thing" in mach3 (one pauses, other
 *		progresses).  sleep() calls setitimer and sigpause.
 *		However, on Mach 2.5, it doesn't work... at least the
 *		sleep never seems to end.  The only problem this might
 *		cause is a stuck c-thread... and that's benign.
 *  4 Oct 92 (mjz): add uid to nameserver name
 */

/***************************************************
 *   Main program for smexec server
 **************************************************/
#include <stdio.h>
#include <mach.h>
#include <mach_error.h>
#include <mig_errors.h>
#include <mach/message.h>
#include <mach/notify.h>
#include <signal.h>
#include <sys/wait.h>
#include <cthreads.h>		/* mach3 bogosity */
#include "smexecd_types.h"
#include "sm.h"			/* SMNAME */
#include <sys/types.h>

extern boolean_t	smexecd_server();	/* MIG supplied */


mach_port_t server_port = MACH_PORT_NULL;
mach_port_t server_port_set = MACH_PORT_NULL;

static char  *ServerName = SMNAME;
int smexecd_debug = 0;

char255 hname;

diesig()
  {
    if (server_port != MACH_PORT_NULL) {
        if (smexecd_debug)
	  printf("gonna netname_check_out!\n");
	netname_check_out(name_server_port,ServerName,MACH_PORT_NULL, 
		server_port);
    }
    exit();
  } /* diesig */

    
chldsig()
{
  union wait status;		/* so we can explain what happend */
  int pid;			/* process id */
  int doneone = 0;

  do {				/* in case > 1 kid died  */
    pid = wait3(&status, WNOHANG,(struct rusage *)0);
    if (pid>0) {			/* was a child to reap */
      fprintf(stderr,"Child %d died, rc=%d, coredump=%d, termsig=%d\n",
	      pid, status.w_retcode, status.w_coredump, status.w_termsig);
      doneone=1;
    } else {
      if (!doneone)
	fprintf(stderr,"Got SIGCHLD, but no child died!?!\n");
    }
  } while (pid > 0);

  return;
} /*chldsig */



/* mach 3 bogosity handler */
int
mach3bogus(arg)
     any_t arg;
{
  while(1) {			/* my favorite kind of loop! */
    sigpause(0);		/* wait around in kernel */
#ifdef VERBOSITY
    fprintf(stderr,"smexecd: sigpause terminated.\n"); /* debug */
#endif VERBOSITY
  }
}

/*
 * To do: add support for port notification
 */

main(argc, argv)
     int argc;
     char *argv[];
{
   kern_return_t krc;
   extern char* smexec_version_string;
   int c;			/* for getopt */
   extern int	optind;		/* for getopt */
   extern char *optarg;		/* for getopt */
   uid_t	uid;
   int		noUid = 0;
   
   while ((c = getopt(argc, argv, "dhqU")) != EOF) {
     switch(c) {
     case 'd':
       smexecd_debug = 1;	/* turn on chattyness - one level for now */
       break;
     case 'q':
       smexecd_debug = 0;	/* turn off; JustInCase default is on */
       break;
     case 'U':
       noUid = 1;		/* backwards compatibility for hahas */
     case 'h':
     case '?':
     default:
       printf("usage: %s -d|-U|-h|-?\n", argv[0]);
       exit(1);
       /*NOTREACHED*/
       break;
     } /* switch */
   } /* while */

   if (!noUid) {
     /* create name based on uid */
     uid = getuid();
     ServerName = calloc(1,240); /* plenty long enough */
     sprintf(ServerName, "%s.%u", SMNAME, (int)uid);
   }

   if (smexecd_debug)
     fprintf(stderr, "Checking '%s' into name server.\n", ServerName);

   signal(SIGINT,diesig);
   signal(SIGCHLD, chldsig); /* so can clean up after kids */

   gethostname(hname,sizeof(hname)); /* for meaningful error messages */
   cthread_detach(cthread_fork(mach3bogus, 0)); /* launch server */

   printf("%s: [Version %s] up and running... \n\tentering server loop. [^C (SIGINT) to kill cleanly]\n", argv[0], smexec_version_string);


   allocate_server_ports(&server_port, &server_port_set, ServerName);

   krc = mach_msg_server(smexecd_server, 1024, server_port_set);
   if (krc != KERN_SUCCESS)  {
       mach_error("mach_msg_server", krc);
   }
   printf("(* !!!!! SMexec server exited - give it up !!!!! *)\n");
   exit(2);  
}






#ifdef notdef


/*
 * procedure server_run:
 *   Waits for messages to server,
 *   handles them, and replies to sender.
 * Stolen in toto from "random" server mig example.
 */
void server_run ()
{
   typedef int space[1024]; /* Maximum message size */

   typedef struct genericMsg
        {
           msg_header_t 	head;
	   msg_type_t		retcodetype;
	   kern_return_t	return_code;
           space 		body;
        } genericMsg;

   msg_return_t		msgcode;
   boolean_t 		ok;
   genericMsg 		*pInMsg, *pRepMsg;

   pInMsg = (genericMsg *)malloc(sizeof(genericMsg));
   pRepMsg = (genericMsg *)malloc(sizeof(genericMsg));

   while (TRUE)
   {
       pInMsg->head.msg_size = sizeof(genericMsg); /* bytes */
       pInMsg->head.msg_local_port = PORT_DEFAULT;

	/* wait to receive request from client */
       msgcode = msg_receive(&pInMsg->head,MSG_OPTION_NONE,0);
       if (msgcode != RCV_SUCCESS)
           mach_error("error in Receive, message will be ignored.\n",
		(kern_return_t)msgcode);
       else 
       {   if (pInMsg->head.msg_type == MSG_TYPE_EMERGENCY)
           {
              if (pInMsg->head.msg_id == NOTIFY_PORT_DELETED)
		  {  /* probably the death of a client's reply */}
	      else
		printf("Unexpected emergency message received: id is %d\n",
			pInMsg->head.msg_id);
           }
          else  /* normal message */
          {   
		/* call server interface module */
	      ok = smexecd_server((msg_header_t *)pInMsg,
				   (msg_header_t *)pRepMsg);
	      if (pRepMsg->return_code != MIG_NO_REPLY)
	      {  
		  /* sending reply message to client */
              	pRepMsg->head.msg_local_port = pInMsg->head.msg_local_port;
              	pRepMsg->head.msg_remote_port = pInMsg->head.msg_remote_port;
              	msgcode = msg_send(&pRepMsg->head,MSG_OPTION_NONE,0);
              	if ((msgcode != SEND_SUCCESS) && 
		    (msgcode != SEND_INVALID_PORT))
                               /* Probably remote process death */
                    mach_error("error at Send.\n",
				(kern_return_t)msgcode);
	      }
	  } /* normal message */
       } /* of message handling */
     }  /* of main loop */

} /* server_run */


#endif


