/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/* RPC called routines of the SMexec server */
/*
 * 28 Feb 92 (mjz): use netnameserver instead of envmgr; latter is not
 *		officially supported under mach3.0, and the function
 *		we need is all in the name server...
 * 19 Mar 92 (mjz): fixed child argv parsing; we were getting it
 *		wrong if no arguments were specified, and mach3 exec()
 *		coredumped.
 * xx Sep 92 (bnb+mjz): switch to use MiG call to get arguments
 *  4 Oct 92 (mjz): pass incarnation number
 */
#include <mach.h>
#include <mach/message.h>
#include <mach/mig_errors.h>
#include <sys/types.h>
#include <sys/stat.h>			/* so I can 'stat' program */
#include <stdio.h>			/* for stderr */
#include "smexecd_types.h"


extern int errno;			
extern char255 hname;
extern mach_port_t server_port_set;


struct reply_struct {
    mach_port_t ss_port;
    mach_port_t ux_port;
    u_long rhost;
    u_short rport;
    char *hlist;
    unsigned int hlistCnt;
    int hcnt;
    int hindex;
    u_long incarn;
};


kern_return_t
new_reply(replyp, ss_port)
    struct reply_struct **replyp;
    mach_port_t ss_port;
{
    struct reply_struct *rpl;
    kern_return_t krc;
    
    rpl = (struct reply_struct*)malloc(sizeof(struct reply_struct));
    krc = allocate_port_for_nameserver((mach_port_t)rpl, ss_port);
    if (krc == KERN_SUCCESS)  {
	*replyp = rpl;
	return KERN_SUCCESS;
    } else {
	if (krc == KERN_NAME_EXISTS)  {
	    krc = new_reply(replyp, ss_port);
	    free(rpl);
	    return krc;
	} else  {
	    mach_error("new_reply", krc);
	    return krc;
	}
    }
}


destroy_reply(argport, ss_port, idbuf)
    mach_port_t	argport;
    mach_port_t ss_port;
    char *idbuf;
{
    kern_return_t krc;
    char255	smsg;
    
    krc = netname_check_out(name_server_port, idbuf, MACH_PORT_NULL);
    if (krc != KERN_SUCCESS) {
	sprintf(smsg,"Error from netname_check_out of %s: %s (%d)",
		idbuf, mach_error_string(krc), krc);
	ss_msg(ss_port, 0, hname, 0, smsg); /* what if this hangs? *mjz* */
    }
    krc = mach_port_destroy(mach_task_self(),argport);
    if (krc != KERN_SUCCESS) {    
	sprintf(smsg,"port_deallocate of argport: %s (%d)",
		mach_error_string(krc), krc);
	ss_msg(ss_port, 0, hname, 0, smsg);
    }
}


/* 1. Verify can read program
 * 2. fork
 * 3. parent: return (squirrel away server, so can wait?)
 *    child: format arguments properly, and exec program
 *		[close any of stdin,stdout,stderr,...?]
 *		Rest of magic performed by main() of child...
 */

kern_return_t start_proc(serv_port,ss_port,ux_port,rhost,rport,
			 hlist,hlistCnt,hcnt,hindex, 
			 incarn,program,progargs)
    mach_port_t      serv_port;			/* server were running on */
    mach_port_t	ss_port;	/* status server port */
    mach_port_t	ux_port;	/* I/O (UniX)  server port */
    u_long	rhost;		/* IP address of rendesvous host */
    u_short	rport;		/* UDP port of rendesvous host */
    char	*hlist;		/* host list; note was vm_allocate'd */
			/* someone must deallocate */
    u_long	hlistCnt;	/* sizeof(hlist) */
    int	hcnt;		/* number of items in hlist */
    u_long	hindex;		/* index of this host in hlist */
    u_long	incarn;		/* incarnation # of this execution */
    char255	program;			/* program to run */
    char255	progargs;			/* args it should told */
{
    int		rc;				/* generic return code */
    kern_return_t krc;
    int		pid;				/* for fork */
    struct      reply_struct *replyp;
    char	idbuf[256];
    char255	smsg;    

    rc = check_program_status(program);
    if (rc)  {
	return rc;
    }


#ifdef TTTESTING
    fprintf(stderr,"UXPING START");
midwayux_ping(    ux_port  );
    fprintf(stderr,"UXPING DONE");    
#endif /*TTTESTING*/
    
    /* Find ourselves a port whose name is the address of the reply */
    krc = new_reply(&replyp);
    if (krc != KERN_SUCCESS)  {
	fprintf(stderr,"could not allocate reply port in start_proc\n");
	return krc;
    }
    
    /* squirrel away a copy of our state passed by smexecd */
    replyp->ss_port = ss_port;
    replyp->ux_port = ux_port;
    replyp->rhost = rhost;
    replyp->rport = rport;
    replyp->hlist = hlist;
    replyp->hlistCnt = hlistCnt;
    replyp->hcnt = hcnt;
    replyp->hindex = hindex;
    replyp->incarn = incarn;

    sprintf(idbuf,"SMARGS.%d", (int)replyp);

    krc = netname_check_in(name_server_port, idbuf, MACH_PORT_NULL,
			   (mach_port_t)replyp);

    if (krc != KERN_SUCCESS)  {
	sprintf(smsg,"netname_check_in of argport: %s (%d)",
		mach_error_string(krc), krc);
	ss_msg(ss_port, 0, hname, 0, smsg);
	return krc;	
    }

    pid = spawn_child(program, progargs, idbuf);
    if (pid < 0)  {
	return (200 + errno);
    } else if (pid == 0)  {
	/* yikes, what about errors here?  wanna send back to status server */
	fprintf(stderr,"execv returned! program was %s\n",program);
	perror("execv");
	_exit(); /* die without causing parent grief */	
	/* NOT REACHED */
    }
    return(KERN_SUCCESS);
} /* start_proc */


kern_return_t
allocate_port_for_nameserver(port_name, ss_port)
    mach_port_t port_name;
    mach_port_t ss_port; /*  */
{
    kern_return_t krc;
    char255	smsg;
    
    /* try to allocate port to transfer rhost, rport, hname.  Can't
     * figure a way to send in argv; may contain 0's...  can't think
     * of a better way, either.  Bletch.
     */
#ifdef TTTESTING
    fprintf(stderr,"CALLING PORT_ALLOCATE IN allocate_port_for_nameserver\n");
#endif

    krc  = mach_port_allocate_name(mach_task_self(),
				   MACH_PORT_RIGHT_RECEIVE,
				   port_name);
    if (krc == KERN_NAME_EXISTS)  {
	return krc;
    }

    if (krc != KERN_SUCCESS)  {
	sprintf(smsg,"Error from port_allocate of argport: %s (%d)",
		mach_error_string(krc), krc);
	ss_msg(ss_port,0,hname, 0, smsg); /* what if this hangs? *mjz* */
	return krc;
    }

    krc = mach_port_move_member(mach_task_self(), port_name, server_port_set);
    if (krc != KERN_SUCCESS)  {
	sprintf(smsg,"Error from mach_port_move_member: %s (%d)",
		mach_error_string(krc), krc);
	ss_msg(ss_port,0,hname,0, smsg); /* what if this hangs? *mjz* */
	return krc;
    }
}




/*
 * Callback function for child to rpc into parent exec daemon.
 * We've conveniently squirreled away the replies from further up on
 * the call chain.
 */

smexec_getargs(port, ss_port, ux_port, rhost, rport, hlist, hlistCnt, 
	       hcnt, hindex, incarn)
    mach_port_t port;
    mach_port_t *ss_port;
    mach_port_t *ux_port;
    u_long *rhost;
    u_short *rport;
    char **hlist;
    unsigned int *hlistCnt;
    int *hcnt;
    int *hindex;
     u_long *incarn;
{
    char idbuf[256];
    struct reply_struct *reply  = (struct reply_struct*)(port); /* wow */

    *ss_port = reply->ss_port;
    *ux_port = reply->ux_port;
    *rhost = reply->rhost;
    *rport = reply->rport;
    *hlist = reply->hlist;
    *hlistCnt = reply->hlistCnt;
    *hcnt = reply->hcnt;
    *hindex = reply->hindex;
    *incarn = reply->incarn;
    
    /*
     * disallow more requests on the specified reply port
     */
    sprintf(idbuf,"SMARGS.%d", (int)reply);
    destroy_reply(reply, ss_port, idbuf);
    free(reply);
    return KERN_SUCCESS;
}

    
    
    
    
smexec_ping(port, res)
    mach_port_t port;
    int *res;
{
    *res = port;
    fprintf(stderr,"Pinged\n");
    return KERN_SUCCESS;
}






int
check_program_status(program)
    char *program;
{
    int rc;
    struct stat statbuf;			/* a place for my stat */
    
    rc = stat(program,&statbuf); /* what is this, anyway? */
    if (rc != 0) return (100+errno); /* return code... */

    if ((statbuf.st_mode & S_IFREG) == 0)  { /* not regular file */
	return(1);
    }
    else {
	/*                      u+x         g+x            o+x */
	if((statbuf.st_mode & (S_IEXEC | (S_IEXEC>>3) | (S_IEXEC>>6))) == 0)
	return(2); /* not executable by anyone */
    }
    return 0;
}

/*
 * patch the following to allow you to attach a gdb and debug
 * over the exec...  attach, set m_w_b_e to 0, and continue.
 */
int midway_wait_before_exec = 0;

spawn_child(program, progargs, idbuf)
    char255 program;
    char255 progargs;
    char *idbuf;
{
    char255	smsg;
    int pid;
    
    /* parse blank separated arg list into argv */
    /* ??? do this first?  is fork/exec better than fork/<chomp>/exec? */
    /* or could use vfork, I guess... */
    /* ***MAGIC*** */
    char *argv[100]; /* hope it's not longer :-( */
    int i;
    int inword=0;

    argv[0] = program; /*do the traditional thing (almost)*/

    /* first argument is the name id to be used to get a handle on
     * invocation arguments */
    argv[1] = idbuf;
    i = 1;   /* last valid thing in argv */
    

    while(*progargs) {
	if ((*progargs == '\t')||(*progargs == ' ')) { /* is whitespace */
	    if(inword) {
		inword = 0; /* no longer in word */
		*progargs = '\0'; /* terminate word */
	    }
	} else {
	    if (!inword) {
		if (i>=98) /* too full */
		break;
		i++; /* advance to next */
		inword = 1; /* starting a word */
		argv[i] = progargs; /* note where */
	    }
	}
	progargs++; /* look at next char */
    } /* while *progargs */
    /* if i 98 either finished or overflowed.  perhaps move this
     * forward so if overflowed, you know it and can return error
     * to caller.  Here, we just press on.
     * Since 98 is max above, we have room for the fence @99
     */
    i++; /* so we don't stomp on the last arg */
    argv[i] = (char *) 0; /* fence */


    /* spawn off the child proc */
    
    pid = fork();
    if (pid == 0)  {
	fprintf(stderr,
		"start_proc: CHILD [%u]: about to call execv\n",
		getpid()); /***/
#ifdef DEBUG_EXEC
	{
	  char255 smsg;
	  int j;
	  sprintf(smsg,
		"[%u] i=%d, ", getpid(),i);
	  for (j=0; j<i; j++)
	    strcat(smsg, argv[j]);
	  fprintf(stderr, "%s\n", smsg);
	}
#endif /* DEBUG_EXEC */
	while (midway_wait_before_exec) cthread_yield();
	execv(program,argv); /* here goes nothing!!! */
	return 0;
    } else  {
	return pid;
    }
    
}   

