/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * Status Server for Shared Memory Exec server
 *
 * 14 Dec 90: Matthew J. Zekauskas
 * 21 Jun 91 (mjz): add host number
 * 29 Jun 92 (mjz): add ss_threat_debug
 * 16 Nov 92 (mjz): add midway_exit() and midway_waitfor_exit()
 * 18 Nov 92 (mjz): now seem to be losing messages.  Add debug gook.
 */

#include <stdio.h>
#include <mach.h>
#include <mach_error.h>
#include <mig_errors.h>
#include <mach/message.h>
#include <mach/notify.h>	/* for NOTIFY_PORT_DELETED */
#include <strings.h>
#include <ss.h>			/* for char255 */
#include <cthreads.h>

extern boolean_t	ss_server();	/* MIG supplied */

static struct mutex ss_stderr_mutex = MUTEX_INITIALIZER;
mutex_t midway_ss_stderr_lock = &ss_stderr_mutex;
int midway_ss_msg_calls = 0;


/*
 * RPC called status server routine (mabye someday routines)
 */
kern_return_t ss_msg(serv_port,loctime,hname,idx,msg)
     mach_port_t	serv_port;	/* server we are running on */
     long	loctime;	/* local "time" -- via Lamport clock */
     char255	hname;		/* source host name */
     long idx;			/* host number (really unsigned)*/
     char255	msg;		/* the message to print */
{
  char *cp;
  char thetime[26];
  extern char* ctime();		/* why isn't this in a header file? */
  long clk;

  /* chop host name at first dot, if there is one */
  if (cp = index(hname,'.')) {
    *cp = '\0';
  }

  mutex_lock(midway_ss_stderr_lock);
  midway_ss_msg_calls++;
  /* get current time.  Danger if ctime called by more than one thread,
   * I imagine. */
  clk = time(0);
  cp = ctime(&clk);		/* Sat_Dec_15_12:43:40_1990\n\0 */
  strncpy(thetime,cp+4,15);	/* take only date and time */
  thetime[15] = '\0';		/* null terminate */

  fprintf(stderr,"-->%s-%s:%u-%8ld: %s\n",thetime,hname,
	  (unsigned) idx,loctime,msg);
  mutex_unlock(midway_ss_stderr_lock);

  return(KERN_SUCCESS);
} /* ss_msg */

/*
 * threat debug routine
 */
int ss_threat_value = 0;
/* keep who did last operation */
long hlast=0,tlast=0;
int  vlast=0,olast=0;
/* if error, move here so not disturbed. */
long herr=0,terr=0;
int  verr=0,oerr=0;

kern_return_t ss_threat_debug(serv_port, hnum, tnum, value, op)
     mach_port_t	serv_port;
     long	hnum;		/* host number */
     long	tnum;		/* thread number */
     int	value;		/* value in question */
     int	op;		/* the operation (really amt to add!) */
{
  char xxx[300];
  if (value != ss_threat_value) {
    /* the dreaded mismatch.  Why out of synch? */
    fprintf(stderr,"=====> threat value mismatch, proc %d thread %d: server (%d) != midway run time (%d), not replying to client thread\n",
	    hnum, tnum, ss_threat_value, value);
    fprintf(stderr,"=====> last was proc %d thread %d op %d\n",
	    hlast, tlast, olast);
    herr = hlast; terr = tlast; verr = vlast; oerr = olast; /* really save */
    /*---warning: mattz below---*/
    sprintf(xxx,"zwrite mattz -m '=====> threat value mismatch, proc %d thread %d: server (%d) != midway run time (%d), not replying to client thread'",
	    hnum, tnum, ss_threat_value, value);
    system(xxx);
    return(MIG_NO_REPLY);
  }

  /* we're OK: do the operation */
  ss_threat_value += op;

  if (op != 0) {
    hlast = hnum; tlast = tnum; vlast = value; olast = op; /* save */
  }

  return(KERN_SUCCESS);
} /* ss_threat_debug */

/*
 * waitfor exit
 */
static struct mutex waitfor_mutex = MUTEX_INITIALIZER;
static mutex_t w_mu = &waitfor_mutex;
static struct condition waitfor_condition = CONDITION_INITIALIZER;
static condition_t w_c = &waitfor_condition;
static int waitfor_item = 1;	/* wait for this to go to zero */

kern_return_t ss_waitfor_exit(serv_port)
     mach_port_t	serv_port;
{
  extern mach_port_t status_server_set;
  extern int ss_server();

  /* start new thread, since we are going to wait */
  cthread_detach(spawn_server_thread(ss_server, status_server_set, 4096));

  mutex_lock(w_mu);
  while(waitfor_item)
    condition_wait(w_c, w_mu);
  mutex_unlock(w_mu);
  return(KERN_SUCCESS);
}

kern_return_t ss_exit(serv_port)
     mach_port_t	serv_port;
{
  waitfor_item = 0;		/* don't worry about protection right now */
  /* (mutex just to give condition something to wait on) */
  condition_broadcast(w_c);	/* let everyone go */
  return (KERN_SUCCESS);	/* done */
}


/* see if we can tell why we are "losing" printf messages. Sigh. */
kern_return_t ss_stats(serv_port)
     mach_port_t	serv_port;
{
  mutex_lock(midway_ss_stderr_lock);
  fprintf(stderr,"ss_stats: %d ss_msg calls.", midway_ss_msg_calls);
  mutex_unlock(midway_ss_stderr_lock);
  return(KERN_SUCCESS);
}

